package admin

import (
	"github.com/graphql-go/graphql"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	q "gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/middleware"
	maintenance "gitlab.mashroom.com/mashroom/backend/maintenance/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	tokenio "gitlab.mashroom.com/mashroom/backend/tokenio/api"
)

const (
	MediaCacheDirEnv       = "MEDIA_CACHE_DIR"
	MediaCacheDirDefault   = "/media"
	SitemapCacheDirEnv     = "SITEMAP_CACHE_DIR"
	SitemapCacheDirDefault = "/sitemaps"
	correlationTypeOffer   = "offer"
)

var (
	MediaCacheDir   = env.ResolveString(MediaCacheDirEnv, MediaCacheDirDefault)
	SitemapCacheDir = env.ResolveString(SitemapCacheDirEnv, SitemapCacheDirDefault)
)

type allParamsPagination struct {
	Page    int32 `json:"page,omitempty"`
	PerPage int32 `json:"perPage,omitempty"`
}

type allParams struct {
	Pagination allParamsPagination    `json:"pagination,omitempty"`
	Sort       []*q.QueryParamsSort   `json:"sort,omitempty"`
	Filter     []*q.QueryParamsFilter `json:"filter,omitempty"`
}

type getParams struct {
	ID string `json:"id"`
}

func InitSchema(rootQueries, rootMutations *graphql.Fields) {
	queryFields := graphql.Fields{}
	mutationFields := graphql.Fields{}

	initTokenioEnrichment()

	initUser(&queryFields, &mutationFields)
	initAdmin(&queryFields, &mutationFields)
	initProperty(&queryFields, &mutationFields)
	initOffer(&queryFields, &mutationFields)
	initDocument(&queryFields, &mutationFields)
	initRightmove(&queryFields, &mutationFields)
	initZoopla(&queryFields, &mutationFields)
	initContract(&queryFields, &mutationFields)
	initCardPayment(&queryFields, &mutationFields)
	initInvoice(&queryFields, &mutationFields)
	initSignature(&queryFields, &mutationFields)
	initMaintenance(&queryFields, &mutationFields)
	initInvitation(&queryFields, &mutationFields)
	initPayment(&queryFields, &mutationFields)
	initGeo(&queryFields, &mutationFields)
	initServices(&queryFields, &mutationFields)
	initStats(&queryFields, &mutationFields)
	initOrder(&queryFields, &mutationFields)
	initMortgage(&queryFields, &mutationFields)
	initNotes(&queryFields, &mutationFields)

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "AdminQuery",
		Description: "Admin query root",
		Fields:      queryFields,
	}))

	(*rootQueries)["admin"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "AdminMutation",
		Description: "Admin mutation root",
		Fields:      mutationFields,
	}))

	(*rootMutations)["admin"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}

	initOfferEnrichment()
	initMaintenanceEnrichment()
}

func initTokenioEnrichment() {
	util.BindOutputName(PaymentUpdate{}, "", util.EnrichmentMap{}, "tokenio_api_PaymentUpdate")
	util.BindOutputName(UpdateSchedule{}, "", util.EnrichmentMap{}, "tokenio_api_UpdateSchedule")

	util.BindOutputName(tokenio.Payment{}, "", util.EnrichmentMap{
		"offer": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use correlation_id / correlation_type",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok || req.CorrelationType != correlationTypeOffer {
					return "", nil
				}

				return req.CorrelationId, nil
			},
		},
		"receipt_id": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use payment_id",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok {
					return "", nil
				}

				return req.PaymentId, nil
			},
		},
	}, "tokenio_api_Payment")

	util.BindOutputName(tokenio.Payment{}, "", util.EnrichmentMap{
		"offer": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use correlation_id / correlation_type",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok || req.CorrelationType != correlationTypeOffer {
					return "", nil
				}

				return req.CorrelationId, nil
			},
		},
		"receipt_id": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use payment_id",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Payment)
				if !ok {
					return "", nil
				}

				return req.PaymentId, nil
			},
		},
	}, "tokenio_api_Payment")

	util.BindOutputName(tokenio.Schedule{}, "", util.EnrichmentMap{
		"offer": &util.Enrichment{
			OutputNullable:    false,
			DeprecationReason: "use correlation_id / correlation_type",
			Output: func(bind util.BindType) graphql.Output {
				return graphql.String
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (interface{}, error) {
				req, ok := p.Source.(*tokenio.Schedule)
				if !ok || req.CorrelationType != correlationTypeOffer {
					return "", nil
				}

				return req.CorrelationId, nil
			},
		},
	}, "tokenio_api_Schedule")
}

func initOfferEnrichment() {
	util.BindOutput(offerData{}, "offer", util.EnrichmentMap{
		"viewing": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.ViewRequest{}, "viewing", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)
				if req.Tenants == nil {
					req.Tenants = &offer.TenantArrayValue{}
				}
				if !ok || len(req.GetTenants().GetValue()) < 1 {
					return &property.ViewRequest{}, nil
				}
				view, err := connections.PropertyService.GetView(p.Context, &property.GetViewRequest{
					User:     req.GetTenants().GetValue()[0].GetUser(),
					Property: req.GetProperty(),
					Id:       req.GetViewing(),
				})
				if err != nil || view == nil || view.GetRequest() == nil {
					return &property.ViewRequest{}, nil
				}
				return view.GetRequest(), nil
			},
		},
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "property", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*offer.Offer)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetProperty() == "" {
					return &property.Property{}, nil
				}

				resp, err := connections.PropertyService.Get(p.Context, &property.GetRequest{Id: req.GetProperty()})
				if err != nil || resp == nil || resp.GetProperty() == nil {
					return &property.Property{}, nil
				}
				return resp.GetProperty(), nil
			},
		},
	})
}

func CensoredAccount(acc *account.Account) *account.Account {
	return &account.Account{
		Id:        acc.GetId(),
		CreatedAt: acc.GetCreatedAt(),
		Details:   acc.GetDetails(),
		Contact:   acc.GetContact(),
	}
}

func ForeignMaintenance(user *auth.AuthResponse, mnt *maintenance.Maintenance) bool {
	return user == nil || (mnt.GetLandlord() != user.GetAccount().GetIdValue())
}

func initMaintenanceEnrichment() {
	util.BindOutput(maintenance.Maintenance{}, "maintenance", util.EnrichmentMap{
		"specialist_someone_else": &util.Enrichment{
			DeprecationReason: "Use specialist",
		},
		"specialist_local_heroes": &util.Enrichment{
			DeprecationReason: "Use specialist",
		},
		"specialist_type": &util.Enrichment{
			DeprecationReason: "Use specialist",
		},
		"occurring_date": &util.Enrichment{
			DeprecationReason: "Use desired_repair_date",
		},
		"category": &util.Enrichment{
			DeprecationReason: "Use taxonomy_id",
		},
		"subcategory": &util.Enrichment{
			DeprecationReason: "Use taxonomy_id",
		},
		"job_description": &util.Enrichment{
			DeprecationReason: "Use title/description",
		},
		"user": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				mnt, ok := p.Source.(*maintenance.Maintenance)
				if !ok || mnt.GetUser() == "" {
					return &account.Account{}, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: mnt.GetUser()})
				if err != nil {
					log.WithError(err).Error()
					return &account.Account{}, nil
				}
				return CensoredAccount(acc.GetAccount()), nil
			},
		},
		"landlord": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&account.Account{}, "", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				mnt, ok := p.Source.(*maintenance.Maintenance)
				if !ok || mnt.GetLandlord() == "" {
					return nil, nil
				}

				acc, err := connections.AccountService.Get(p.Context, &account.GetRequest{Id: mnt.GetLandlord()})
				if err != nil {
					log.WithError(err).Error()
					return nil, nil
				}

				user, ok := p.Context.Value(ws.CtxUser).(*auth.AuthResponse)
				if !ok || user == nil || ForeignMaintenance(user, mnt) {
					return CensoredAccount(acc.GetAccount()), nil
				}
				return acc.GetAccount(), err
			},
		},
		"property": &util.Enrichment{
			OutputNullable: true,
			Output: func(bind util.BindType) graphql.Output {
				return util.BindOutput(&property.Property{}, "property", nil)
			},
			Resolver: func(p graphql.ResolveParams, oldValue interface{}) (i interface{}, e error) {
				req, ok := p.Source.(*maintenance.Maintenance)
				if !ok {
					return &property.Property{}, nil
				}

				if req.GetProperty() == "" {
					return &property.Property{}, nil
				}

				resp, err := connections.PropertyService.Get(
					p.Context,
					&property.GetRequest{
						Id: req.GetProperty(),
					},
				)
				if err != nil || resp == nil || resp.GetProperty() == nil {
					return &property.Property{}, nil
				}
				return resp.GetProperty(), nil
			},
		},
	})
}
