package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	cardpayment "gitlab.mashroom.com/mashroom/backend/cardpayment/api"
	q "gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
)

type confirmCardpaymentRequest struct {
	ChargeID string `json:"charge_id"`
	ID       string `json:"id"`
	Confirm  bool   `json:"confirm"`
}

type allCardpaymentsResult struct {
	Data         []*cardpayment.ReceiptItem `json:"data,omitempty"`
	Total        int32                      `json:"total,omitempty"`
	NonConfirmed int32                      `json:"non_confirmed,omitempty"`
}

func initCardPayment(query, mutation *graphql.Fields) {
	(*query)["Cardpayment"] = &graphql.Field{
		DeprecationReason: "use Invoice",
		Description:       "Retrieves specified cardpayment item",
		Args:              util.BindArguments(cardpayment.ReceiptItemRequest{}),
		Resolve:           adminQueryGetCardpaymentResolver,
		Type:              util.BindOutput(cardpayment.ReceiptItem{}, "Receipt item information", nil),
	}
	(*query)["allCardpayments"] = &graphql.Field{
		DeprecationReason: "use allInvoices",
		Args:              util.BindArguments(allParams{}),
		Description:       "Get all receipts",
		Resolve:           adminQueryAllCardpaymentsResolver,
		Type:              util.BindOutput(allCardpaymentsResult{}, "Receipt items list", nil),
	}

	(*mutation)["confirmCardpayment"] = &graphql.Field{
		DeprecationReason: "use updateInvoiceItems",
		Description:       "Confirm a receipt item",
		Args:              util.BindArguments(confirmCardpaymentRequest{}),
		Resolve:           adminMutationConfirmCardpaymentResolver,
		Type:              util.BindOutput(cardpayment.ReceiptItem{}, "Receipt item information", nil),
	}
}

func renderReceiptItem(
	inv,
	ref,
	user,
	property string,
	pstatus invoice.PaymentStatus,
	stripe *invoice.PaymentSystemStripe,
	service *extras.ServiceSnapshotStatus,
) *cardpayment.ReceiptItem {
	var status cardpayment.PaymentStatus

	switch pstatus {
	case invoice.PaymentStatus_PAYMENT_SUCCESS:
		status = cardpayment.PaymentStatus_SUCCEEDED
	case invoice.PaymentStatus_PAYMENT_PENDING:
		status = cardpayment.PaymentStatus_PROCESSING
	case invoice.PaymentStatus_PAYMENT_CONFIRM:
		status = cardpayment.PaymentStatus_PROCESSING
	case invoice.PaymentStatus_PAYMENT_FAILED:
		status = cardpayment.PaymentStatus_FAILED
	}

	return &cardpayment.ReceiptItem{
		ChargeId:      inv,
		ChargeType:    stripe.GetRequest().GetType(),
		ClientSecret:  stripe.GetResponse().GetClientSecret(),
		UserId:        user,
		UserEmail:     "",
		Property:      property,
		ChargeStatus:  status,
		CreatedAt:     service.GetService().GetCreated(),
		UpdatedAt:     service.GetUpdatedAt(),
		InvoiceNumber: ref,
		Admin:         service.GetAdmin(),
		ConfirmedAt:   service.GetUpdatedAt(),
		Confirmed:     service.GetStatus() == extras.ServiceStatus_ACTIVE,
		Id: &cardpayment.ReceiptItem_IdValue{
			IdValue: service.GetService().GetId(),
		},
		ServiceId: &cardpayment.ReceiptItem_ServiceIdValue{
			ServiceIdValue: service.GetService().GetServiceId(),
		},
		Code: &cardpayment.ReceiptItem_CodeValue{
			CodeValue: service.GetService().GetCode(),
		},
		Type: &cardpayment.ReceiptItem_TypeValue{
			TypeValue: service.GetService().GetType(),
		},
		Status: &cardpayment.ReceiptItem_StatusValue{
			StatusValue: service.GetStatus(),
		},
		Role: &cardpayment.ReceiptItem_RoleValue{
			RoleValue: service.GetService().GetRole(),
		},
		Price: &cardpayment.ReceiptItem_PriceValue{
			PriceValue: service.GetService().GetPrice(),
		},
		Description: &cardpayment.ReceiptItem_DescriptionValue{
			DescriptionValue: service.GetService().GetDescription(),
		},
		Primary: &cardpayment.ReceiptItem_PrimaryValue{
			PrimaryValue: service.GetService().GetPrimary(),
		},
	}
}

func renderRceiptInvoice(inv *invoice.Invoice, id string) *cardpayment.ReceiptItem {
	var payment *invoice.Payment

	for _, p := range inv.GetPayments() {
		if p.GetStatus() == invoice.PaymentStatus_PAYMENT_CANCELED || p.GetSystem().GetStripe() == nil {
			continue
		}

		if payment == nil || p.GetCreatedAt().AsTime().After(payment.GetCreatedAt().AsTime()) {
			payment = p
		}
	}

	for _, p := range inv.GetServices() {
		if p.GetService().GetId() == id {
			return renderReceiptItem(
				inv.GetInvoiceId(),
				inv.GetReference(),
				inv.GetUser(),
				inv.GetProperty(),
				payment.GetStatus(),
				payment.GetSystem().GetStripe(),
				p,
			)
		}
	}

	return nil
}

func adminQueryGetCardpaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &cardpayment.ReceiptItemRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.InvoiceService.InvoiceGet(params.Context, &invoice.InvoiceGetRequest{
		InvoiceId: request.GetChargeId(),
	})
	if err != nil {
		return nil, err
	}

	rcpt := renderRceiptInvoice(resp.GetInvoice(), request.GetId())
	if rcpt == nil {
		return nil, grpcerr.NotFound
	}

	return rcpt, nil
}

func adminQueryAllCardpaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	// nolint:staticcheck // TemporaryItemFilterList and TemporaryItemFilterListRequest are DEPRECATED
	resp, err := connections.InvoiceService.TemporaryItemFilterList(
		params.Context,
		&invoice.TemporaryItemFilterListRequest{
			Params: req,
		},
	)
	if err != nil {
		return nil, err
	}

	var items []*cardpayment.ReceiptItem

	for _, it := range resp.GetItems() {
		items = append(items, renderReceiptItem(
			it.GetInvoiceId(),
			it.GetInvoiceReference(),
			it.GetUser(),
			it.GetProperty(),
			it.GetPaymentStatus(),
			it.GetStripe(),
			it.GetService(),
		))
	}

	return allCardpaymentsResult{
		Data:  items,
		Total: resp.GetTotal(),
	}, nil
}

func adminMutationConfirmCardpaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &confirmCardpaymentRequest{}

	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	var status extras.ServiceStatus

	if req.Confirm {
		status = extras.ServiceStatus_ACTIVE
	} else {
		status = extras.ServiceStatus_PENDING
	}

	resp, err := connections.InvoiceService.ServiceItemStatus(params.Context, &invoice.ServiceItemStatusRequest{
		InvoiceId: req.ChargeID,
		AdminId:   user.GetAccount().GetIdValue(),
		Statuses: []*invoice.ServiceItemStatus{
			{
				ItemId: req.ID,
				Status: status,
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return renderRceiptInvoice(resp.GetInvoice(), req.ID), nil
}
