package admin

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	q "gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers"
	mortgage "gitlab.mashroom.com/mashroom/backend/mortgage/api"
)

func initMortgage(query, mutation *graphql.Fields) {
	(*query)["mortgageAllBookings"] = &graphql.Field{
		Description: "List all mortgage bookings",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllBookingsResolver,
		Type:        util.BindOutput(mortgage.BookingFilterListResponse{}, "Booking list", nil),
	}
	(*query)["mortgageAllLeads"] = &graphql.Field{
		Description: "List all mortgage leads",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllMortgageLeadsResolver,
		Type:        util.BindOutput(mortgage.LeadListResponse{}, "Leads list", nil),
	}

	(*mutation)["mortgageLenderSync"] = &graphql.Field{
		Description: "Sync lender list",
		Args:        util.BindArguments(mortgage.ServiceLenderSyncRequest{}),
		Resolve:     adminMutationLenderSyncResolver,
		Type:        graphql.Boolean,
	}
	(*mutation)["saveMortgageLead"] = &graphql.Field{
		Description: "Save mortgage lead",
		Args:        util.BindArguments(mortgage.LeadSaveRequest{}),
		Resolve:     adminMutationMortgageLeadSaveResolver,
		Type:        util.BindOutput(mortgage.LeadSaveResponse{}, "Saved lead", nil),
	}
	(*mutation)["saveMortgageLeadDeal"] = &graphql.Field{
		Description: "Save mortgage lead deal",
		Args:        util.BindArguments(mortgage.LeadDealSaveRequest{}),
		Resolve:     adminMutationMortgageLeadDealSaveResolver,
		Type:        util.BindOutput(mortgage.LeadDealSaveResponse{}, "Saved lead deal", nil),
	}
	(*mutation)["saveMortgageLeadDealNote"] = &graphql.Field{
		Description: "Save mortgage lead deal note",
		Args:        util.BindArguments(mortgage.LeadDealNoteSaveRequest{}),
		Resolve:     adminMutationMortgageLeadDealNoteSaveResolver,
		Type:        util.BindOutput(mortgage.LeadDealNoteSaveResponse{}, "Saved lead deal note", nil),
	}
	(*mutation)["saveMortgageLeadDocument"] = &graphql.Field{
		Description: "Save mortgage lead document",
		Args:        util.BindArguments(mortgage.LeadDocumentSaveRequest{}),
		Resolve:     adminMutationMortgageLeadDocumentSaveResolver,
		Type:        util.BindOutput(mortgage.LeadDocumentSaveResponse{}, "Saved lead document", nil),
	}
}

func adminQueryAllBookingsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created", Order: -1}}
	}

	return connections.MortgageService.BookingFilterList(params.Context, &mortgage.BookingFilterListRequest{
		Params: req,
	})
}

func adminQueryAllMortgageLeadsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created", Order: -1}}
	}

	return connections.MortgageService.LeadList(params.Context, &mortgage.LeadListRequest{
		Params: req,
	})
}

func adminMutationLenderSyncResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &mortgage.ServiceLenderSyncRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	_, err := connections.MortgageService.ServiceLenderSync(params.Context, req)
	if err != nil {
		return nil, err
	}

	return true, nil
}

func adminMutationMortgageLeadSaveResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &mortgage.LeadSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	req.AdminId = user.GetAccount().GetIdValue()

	return connections.MortgageService.LeadSave(params.Context, req)
}

func adminMutationMortgageLeadDealSaveResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &mortgage.LeadDealSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.MortgageService.LeadDealSave(params.Context, req)
}

func adminMutationMortgageLeadDealNoteSaveResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &mortgage.LeadDealNoteSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.MortgageService.LeadDealNoteSave(params.Context, req)
}

func adminMutationMortgageLeadDocumentSaveResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &mortgage.LeadDocumentSaveRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.MortgageService.LeadDocumentSave(params.Context, req)
}
