package admin

import (
	"context"

	"github.com/graphql-go/graphql"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	q "gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/wrap"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
)

type allOffersResult struct {
	Data  []*offer.Offer `json:"data,omitempty"`
	Total int32          `json:"total,omitempty"`
}

type allScheduledPaymentsResult struct {
	Data  []*offer.ScheduledPayment `json:"data,omitempty"`
	Total int32                     `json:"total,omitempty"`
}

type OfferFull struct {
	*offer.Offer
	Viewing  *property.ViewRequest `json:"viewing,omitempty"`
	Property *property.Property    `json:"property,omitempty"`
}

func (f OfferFull) ProvideOffer() *offer.Offer {
	return f.Offer
}

type offerData struct {
	OfferFull
	Offers   []*offer.Offer          `json:"offers,omitempty"`
	Viewings []*property.ViewRequest `json:"viewings,omitempty"`
}

func (p offerData) ProvideOffer() *offer.Offer {
	return p.Offer
}

type offerUpdateResponse struct {
	Status offer.OfferStatus  `json:"status,omitempty"`
	Errors []offer.OfferError `json:"errors,omitempty"`
	Offer  *offerData         `json:"offer,omitempty"`
}

var (
	errPropertyNotFound = grpcerr.New(grpcerr.InvalidData, "property not found")
	errLandlordNotFound = grpcerr.New(grpcerr.InvalidData, "landlord not found")
)

type allParamsOffer struct {
	Pagination allParamsPagination    `json:"pagination,omitempty"`
	Sort       []*q.QueryParamsSort   `json:"sort,omitempty"`
	Filter     []*q.QueryParamsFilter `json:"filter,omitempty"`
}

func initOffer(query, mutation *graphql.Fields) {
	(*query)["Offer"] = &graphql.Field{
		Description: "Retrieves specified offer",
		Args:        util.BindArguments(offer.GetRequest{}),
		Resolve:     adminQueryGetOfferResolver,
		Type:        util.BindOutput(offerData{}, "Offer details", nil),
	}
	(*query)["allOffers"] = &graphql.Field{
		Description: "Retrieves offer list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllOffersResolver,
		Type:        util.BindOutput(allOffersResult{}, "Offer list", nil),
	}
	(*query)["allAdvancedPayments"] = &graphql.Field{
		DeprecationReason: "use allScheduledPayments",
		Description:       "Retrieves advance payments list",
		Args:              util.BindArguments(allParams{}),
		Resolve:           adminQueryAllAdvancePaymentsResolver,
		Type:              util.BindOutput(allScheduledPaymentsResult{}, "Scheduled payments list", nil),
	}
	(*query)["allScheduledPayments"] = &graphql.Field{
		Description: "Retrieves scheduled payments list",
		Args:        util.BindArguments(allParamsOffer{}),
		Resolve:     adminQueryAllScheduledPaymentsResolver,
		Type:        util.BindOutput(allScheduledPaymentsResult{}, "Scheduled payments list", nil),
	}

	(*mutation)["updateOffer"] = &graphql.Field{
		Description: "Updates existing offer",
		Args:        util.BindArguments(offer.ServiceUpdateRequest{}),
		Resolve:     adminMutationUpdateOfferResolver,
		Type:        util.BindOutput(offerUpdateResponse{}, "Update status", nil),
	}

	(*mutation)["acceptOffer"] = &graphql.Field{
		Description: "Accepts offer on behalf of a Landlord",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationAcceptOfferResolver,
		Type:        util.BindOutput(offer.AcceptResponse{}, "Accept status", nil),
	}

	(*mutation)["approveOffer"] = &graphql.Field{
		Description: "Approves prepared offer",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationApproveOfferResolver,
		Type:        util.BindOutput(offer.ApproveResponse{}, "Approve status", nil),
	}

	(*mutation)["OfferPaymentStatus"] = &graphql.Field{
		DeprecationReason: "Use updatePaymentStatus",
		Description:       "Updates (confirms) offer payment by tenant via bank",
		Args:              util.BindArguments(offer.PaymentStatusRequest{}),
		Resolve:           adminMutationOfferPaymentStatusResolver,
		Type:              util.BindOutput(offer.PaymentStatusResponse{}, "Offer payment status", nil),
	}

	(*mutation)["RenderOfferCsv"] = &graphql.Field{
		Description: "Generate csv with Tenant and Landlord details to send to the insurer",
		Args:        util.BindArguments(offer.RenderCsvRequest{}),
		Resolve:     adminMutationRenderCsvResolver,
		Type:        util.BindOutput(offer.RenderCsvResponse{}, "Render csv response", nil),
	}

	(*mutation)["executeOfferCheck"] = &graphql.Field{
		Description: "Execute tenants check for offer",
		Args:        util.BindArguments(offer.ExecuteCheckRequest{}),
		Resolve:     adminMutationCheckOfferResolver,
		Type:        util.BindOutput(offer.CheckResponse{}, "Check offer tenants response", nil),
	}
	(*mutation)["updateScheduledPayment"] = &graphql.Field{
		Description: "Updates advancement status on scheduled pyament",
		Args:        util.BindArguments(offer.ScheduledPayment{}),
		Resolve:     adminMutationUpdateScheduledPaymentResolver,
		Type:        util.BindOutput(offer.ScheduledPayment{}, "Update advancement status response", nil),
	}
	(*mutation)["updateDraftContract"] = &graphql.Field{
		Description: "Update draft contract. Works only on Contract Review state",
		Args:        util.BindArguments(offer.UpdateDraftContractRequest{}),
		Resolve:     adminMutationUpdateDraftContractResolver,
		Type:        util.BindOutput(offer.UpdateDraftContractResponse{}, "update draft contract response", nil),
	}
	(*mutation)["skipReferencingStage"] = &graphql.Field{
		Description: "Skip offer referencing stage (AWAITING_CHECKS and AWAITING_APPLICANT_CHECK_PAYMENT)",
		Args:        util.BindArguments(offer.SkipReferencingStageRequest{}),
		Resolve:     adminMutationSkipReferencingStageResolver,
		Type:        util.BindOutput(offer.SkipReferencingStageResponse{}, "skip referencing checks response", nil),
	}
	(*mutation)["cancelSigningOffer"] = &graphql.Field{
		Description: "Cancel signing offer",
		Args:        util.BindArguments(offer.CancelSigningStageRequest{}),
		Resolve:     adminMutationCancelSigningOfferResolver,
		Type:        util.BindOutput(offer.CancelSigningStageResponse{}, "Cancel Signing offer response", nil),
	}
	(*mutation)["setRiskReport"] = &graphql.Field{
		Description: "Forces to set Renprofile report for specified offer and tenant or guarantor",
		Args: func() graphql.FieldConfigArgument {
			cfg := util.BindArguments(offer.SetRiskReportRequest{})
			delete(cfg, "sender_id")

			return cfg
		}(),
		Resolve: resolveSetRentprofileReport,
		Type:    graphql.Boolean,
	}
	(*mutation)["setAdminOffer"] = &graphql.Field{
		Description: "Add an admin user to an offer",
		Args:        util.BindArguments(offer.SetAdminRequest{}),
		Resolve:     adminMutationSetAdminOfferResolver,
		Type:        graphql.Boolean,
	}
}

func findLandlord(ctx context.Context, landlord *offer.Landlord) *account.Account {
	req := &q.QueryParams{
		Filter: []*q.QueryParamsFilter{},
		Sort: []*q.QueryParamsSort{{
			Field: "id",
			Order: 1,
		}},
	}
	if (landlord.GetEmailPrimary() != nil) && (landlord.GetEmailPrimary().GetValue() != "") {
		req.Filter = append(req.Filter, &q.QueryParamsFilter{
			Fields:   []string{"contact.email_primary"},
			Value:    []string{landlord.GetEmailPrimary().GetValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}
	if (landlord.GetPhonePrimary() != nil) && (landlord.GetPhonePrimary().GetValue() != "") {
		req.Filter = append(req.Filter, &q.QueryParamsFilter{
			Fields:   []string{"contact.phone_primary"},
			Value:    []string{landlord.GetPhonePrimary().GetValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}
	request := &account.AccountListFilterRequest{
		Params: &account.AccountListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.Admin = false
	resp, err := connections.AccountService.ListFilter(ctx, request)
	if (err != nil) || (len(resp.GetAccounts()) != 1) {
		return nil
	}
	return resp.GetAccounts()[0]
}

func getViewing(ctx context.Context, off *offer.Offer) (*property.ViewRequest, error) {
	if off.GetTenants() == nil {
		off.Tenants = &offer.TenantArrayValue{}
	}
	if len(off.GetTenants().GetValue()) < 1 {
		return &property.ViewRequest{}, nil
	}
	view, err := connections.PropertyService.GetView(ctx, &property.GetViewRequest{
		User:     off.GetTenants().GetValue()[0].GetUser(),
		Property: off.GetProperty(),
		Id:       off.GetViewing(),
	})
	if err != nil || view == nil || view.GetRequest() == nil {
		return &property.ViewRequest{}, nil
	}
	return view.GetRequest(), nil
}

func getProperty(ctx context.Context, off *offer.Offer) (*property.Property, error) {
	if off.GetProperty() == "" {
		return &property.Property{}, nil
	}
	resp, err := connections.PropertyService.Get(ctx, &property.GetRequest{Id: off.GetProperty()})
	if err != nil || resp == nil || resp.GetProperty() == nil {
		return &property.Property{}, nil
	}
	return resp.GetProperty(), nil
}

func enrichOfferData(ctx context.Context, off *offer.Offer, resp *offerData) {
	if off == nil {
		return
	}
	// Get offers for property from this offer (maybe filter by tenant too?)
	req := &q.QueryParams{
		Filter: []*q.QueryParamsFilter{
			{
				Fields:   []string{"property"},
				Value:    []string{off.GetProperty()},
				Operator: q.QueryFilterOperator_EQUALS,
			},
		},
		Sort: []*q.QueryParamsSort{{
			Field: "id",
			Order: 1,
		}},
	}
	request := &offer.ListFilterRequest{
		Params: &offer.ListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.AdminSite = true
	if offs, err := connections.OfferService.ListFilter(ctx, request); err == nil {
		resp.Offers = offs.Offers
	}

	if off.GetLandlord() == nil {
		off.Landlord = &offer.Landlord{}
	}
	// Get view requests list for property from this offer
	landlord := findLandlord(ctx, off.GetLandlord())
	if landlord != nil {
		request := &property.ListViewRequest{
			Property: wrap.StringValue(&off.Property),
			User:     landlord.GetIdValue(),
		}
		viewings, err := connections.PropertyService.ListView(ctx, request)
		if err == nil {
			resp.Viewings = viewings.List
		}
	}

	view, err := getViewing(ctx, off)
	if (err == nil) && (view != nil) {
		resp.Viewing = view
	}

	prop, err := getProperty(ctx, off)
	if (err == nil) && (prop != nil) {
		resp.Property = prop
	}
}

func adminQueryGetOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	offr, err := connections.OfferService.ServiceGet(params.Context, request)
	if err != nil {
		return nil, err
	}
	var resp *offerData = nil
	if offr.GetOffer() != nil {
		resp = &offerData{
			OfferFull: OfferFull{
				Offer: offr.GetOffer(),
			},
		}
	}
	if (offr != nil) && (resp != nil) {
		enrichOfferData(params.Context, offr.GetOffer(), resp)
	}
	return resp, nil
}

func adminQueryAllOffersResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	input := &allParams{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}
	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}
	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created_at", Order: -1}}
	}
	request := &offer.ListFilterRequest{
		Params: &offer.ListFilterRequest_ParamsValue{ParamsValue: req},
	}
	request.AdminSite = true
	resp, err := connections.OfferService.ListFilter(params.Context, request)
	if err != nil {
		return nil, err
	}
	offers := resp.GetOffers()
	total := resp.GetTotal()
	return allOffersResult{
		Data:  offers,
		Total: total,
	}, nil
}

func adminQueryAllAdvancePaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	req.Filter = append(req.Filter, &q.QueryParamsFilter{
		Fields:   []string{"advance.state"},
		Value:    []string{"1", "2", "3"},
		Operator: q.QueryFilterOperator_ANY,
	})

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "date", Order: -1}}
	}

	request := &offer.ScheduledPaymentListRequest{
		Params: req,
	}

	resp, err := connections.OfferService.ScheduledPaymentList(params.Context, request)
	if err != nil {
		return nil, err
	}

	return allScheduledPaymentsResult{
		Data:  resp.GetItems(),
		Total: resp.GetTotal(),
	}, nil
}

func adminQueryAllScheduledPaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParamsOffer{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "date", Order: -1}}
	}

	request := &offer.FilterScheduledPaymentsRequest{
		Params: req,
	}

	resp, err := connections.OfferService.FilterScheduledPayments(params.Context, request)
	if err != nil {
		return nil, err
	}

	return allScheduledPaymentsResult{
		Data:  resp.GetItems(),
		Total: resp.GetTotal(),
	}, nil
}

func adminMutationUpdateOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.ServiceUpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.OfferService.ServiceUpdate(params.Context, request)
	if err != nil {
		return nil, err
	}
	var offr *offerData = nil
	if resp.GetOffer() != nil {
		offr = &offerData{
			OfferFull: OfferFull{
				Offer: resp.GetOffer(),
			},
		}
	}
	if (resp != nil) && (offr != nil) {
		enrichOfferData(params.Context, resp.GetOffer(), offr)
	}
	return &offerUpdateResponse{
		Status: resp.Status,
		Errors: resp.Errors,
		Offer:  offr,
	}, nil
}

func adminMutationAcceptOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	reqOffer := &offer.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, reqOffer); err != nil {
		return nil, err
	}
	offr, err := connections.OfferService.ServiceGet(params.Context, reqOffer)
	if err != nil {
		return nil, err
	}
	if (offr.GetOffer() == nil) || (offr.GetOffer().GetProperty() == "") {
		return nil, errPropertyNotFound
	}

	reqProp := &property.GetRequest{
		Id: offr.GetOffer().GetProperty(),
	}
	prop, err := connections.PropertyService.Get(params.Context, reqProp)
	if err != nil {
		return nil, err
	}
	if (prop.GetProperty() == nil) ||
		(prop.GetProperty().GetPeople() == nil) ||
		(prop.GetProperty().GetPeople().GetLandlord() == nil) ||
		(prop.GetProperty().GetPeople().GetLandlord().GetValue() == "") {
		return nil, errLandlordNotFound
	}
	landlord := prop.GetProperty().GetPeople().GetLandlord().GetValue()

	request := &offer.AcceptRequest{
		User: landlord,
		Id:   reqOffer.Id,
	}
	return connections.OfferService.Accept(params.Context, request)
}

func adminMutationApproveOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &offer.ApproveRequest{
		User:  user.GetAccount().GetIdValue(),
		Offer: req.ID,
	}
	offr, err := connections.OfferService.ServiceApprove(params.Context, request)
	if err != nil {
		return nil, err
	}
	var resp *offerData = nil
	if offr.GetOffer() != nil {
		resp = &offerData{
			OfferFull: OfferFull{
				Offer: offr.GetOffer(),
			},
		}
	}
	if (offr != nil) && (resp != nil) {
		enrichOfferData(params.Context, offr.GetOffer(), resp)
	}
	return resp, nil
}

func adminMutationOfferPaymentStatusResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &offer.PaymentStatusRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	inv, err := connections.InvoiceService.InvoiceGet(params.Context, &invoice.InvoiceGetRequest{
		InvoiceId: request.GetReceiptId(),
	})
	if err != nil {
		return nil, err
	}

	var latest *invoice.Payment

	for _, p := range inv.GetInvoice().GetPayments() {
		if p.GetCreatedAt().AsTime().After(latest.GetCreatedAt().AsTime()) {
			latest = p
		}
	}

	var status invoice.PaymentStatus

	switch request.GetStatus() {
	case offer.PaymentStatus_PAYMENT_SUCCESS:
		status = invoice.PaymentStatus_PAYMENT_SUCCESS
	case offer.PaymentStatus_PAYMENT_PENDING:
		status = invoice.PaymentStatus_PAYMENT_PENDING
	case offer.PaymentStatus_PAYMENT_FAILED:
		status = invoice.PaymentStatus_PAYMENT_FAILED
	case offer.PaymentStatus_PAYMENT_CONFIRM:
		status = invoice.PaymentStatus_PAYMENT_CONFIRM
	}

	resp, err := connections.InvoiceService.ServicePaymentStatus(params.Context, &invoice.ServicePaymentStatusRequest{
		PaymentId:     latest.GetPaymentId(),
		AdminId:       user.GetAccount().GetIdValue(),
		Status:        status,
		Description:   "",
		TransactionId: "",
	})
	if err != nil {
		return nil, err
	}

	var offerID string

	if inv.GetInvoice().GetCorrelationType() == correlationTypeOffer {
		offerID = inv.GetInvoice().GetCorrelationId()
	}

	return &offer.PaymentStatusResponse{
		Ok:        resp.GetOk(),
		Offer:     offerID,
		ReceiptId: inv.GetInvoice().GetInvoiceId(),
	}, nil
}

func adminMutationRenderCsvResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.RenderCsvRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.OfferService.RenderCsv(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func adminMutationCheckOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.ExecuteCheckRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	resp, err := connections.OfferService.ExecuteCheck(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func adminMutationUpdateScheduledPaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	item := &offer.ScheduledPayment{}

	if err := util.UnmarshalArguments(params.Args, item); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	resp, err := connections.OfferService.ScheduledPaymentSave(params.Context, &offer.ScheduledPaymentSaveRequest{
		User: user.GetAccount().GetIdValue(),
		Item: item,
	})
	if err != nil {
		return nil, err
	}

	return resp.GetItem(), nil
}

func adminMutationUpdateDraftContractResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.UpdateDraftContractRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.OfferService.UpdateDraftContract(params.Context, request)
}

func adminMutationSkipReferencingStageResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.SkipReferencingStageRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.OfferService.SkipReferencingStage(params.Context, request)
}

func adminMutationCancelSigningOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &offer.CancelSigningStageRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.OfferService.CancelSigningStage(params.Context, request)

	return resp, err
}

func adminMutationSetAdminOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &offer.SetAdminRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	_, err := connections.OfferService.SetAdmin(params.Context, req)

	return err == nil, err
}

func resolveSetRentprofileReport(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &offer.SetRiskReportRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.SenderId = user.GetAccount().GetIdValue()

	_, err := connections.OfferService.SetRiskReport(params.Context, request)

	return err != nil, err
}
