package admin

import (
	"context"
	"strconv"
	"time"

	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers/errs"

	"github.com/graphql-go/graphql"
	q "gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	tokenio "gitlab.mashroom.com/mashroom/backend/tokenio/api"
)

var errOfferNotFound = grpcerr.New(grpcerr.InvalidData, "offer not fond")

type getPaymentsFilterResult struct {
	Data  []*tokenio.Payment `json:"data,omitempty"`
	Total int32              `json:"total,omitempty"`
}

type Subscribe struct{}
type Unsubscribe struct{}

type MatchPropertyPaymentsAllRequest struct {
	tokenio.MatchPropertyPaymentsAllRequest
	Offer string `json:"offer"`
}

type GetScheduleRequest struct {
	tokenio.GetScheduleRequest
	Offer string `json:"offer"`
}

type MatchPropertyPaymentRequest struct {
	tokenio.MatchPropertyPaymentRequest
	Offer string `json:"offer"`
}

type PaymentUpdate struct {
	tokenio.PaymentUpdate
	Offer     string `json:"offer"`
	ReceiptID string `json:"receipt_id"`
}

type UpdatePaymentRequest struct {
	Payment *PaymentUpdate
}

type UpdateSchedule struct {
	tokenio.UpdateSchedule
	Offer string `json:"offer"`
}

type UpdateScheduleRequest struct {
	Schedule *UpdateSchedule
}

type CreateTransferPayoutRequest struct {
	tokenio.CreateTransferPayoutRequest
	Offer    string `json:"offer"`
	Property string `json:"property"`
}

func initPayment(query, mutation *graphql.Fields) {
	(*query)["getPayments"] = &graphql.Field{
		DeprecationReason: "use Invoice",
		Description:       "Retrieves payments list",
		Args:              util.BindArguments(tokenio.GetPaymentsRequest{}),
		Resolve:           adminQueryGetPaymentsResolver,
		Type:              util.BindOutput(tokenio.GetPaymentsResponse{}, "Get payments", nil),
	}
	(*query)["getPaymentsFilter"] = &graphql.Field{
		DeprecationReason: "use allInvoices",
		Description:       "Retrieves payments list with sort and pagination",
		Args:              util.BindArguments(allParams{}),
		Resolve:           adminQueryGetPaymentsFilterResolver,
		Type:              util.BindOutput(getPaymentsFilterResult{}, "Get payments", nil),
	}
	(*query)["getSchedule"] = &graphql.Field{
		DeprecationReason: "use allScheduledPayments",
		Description:       "Get a schedule",
		Args:              util.BindArguments(GetScheduleRequest{}),
		Resolve:           adminQueryGetScheduleResolver,
		Type:              util.BindOutput(tokenio.GetScheduleResponse{}, "Get schedule by offer or property", nil),
	}
	(*query)["matchPayments"] = &graphql.Field{
		DeprecationReason: "use matchTransactions",
		Description:       "An attempt to match payments to the current payment amount for offer/property",
		Args:              util.BindArguments(MatchPropertyPaymentRequest{}),
		Resolve:           adminQueryMatchPaymentsResolver,
		Type: util.BindOutput(
			tokenio.MatchPropertyPaymentResponse{},
			"Get payments for current payment amount",
			nil,
		),
	}

	(*mutation)["matchAndUpdatePayments"] = &graphql.Field{
		DeprecationReason: "use matchTransactions",
		Description: "An attempt to match all possible payments for offer/property" +
			" and update schedule in success cases",
		Args:    util.BindArguments(MatchPropertyPaymentsAllRequest{}),
		Resolve: adminMutationMatchPaymentsResolver,
		Type:    util.BindOutput(tokenio.MatchPropertyPaymentsAllResponse{}, "Get payments for matched cases", nil),
	}
	(*mutation)["updatePayment"] = &graphql.Field{
		DeprecationReason: "use updateInvoiceItems",
		Description:       "Download payments from tokenio",
		Args:              util.BindArguments(UpdatePaymentRequest{}),
		Resolve:           adminMutationUpdatePaymentResolver,
		Type:              util.BindOutput(tokenio.UpdatePaymentResponse{}, "Update payment", nil),
	}
	(*mutation)["updateSchedule"] = &graphql.Field{
		DeprecationReason: "use updateScheduledPayment",
		Description:       "Update a schedule",
		Args:              util.BindArguments(UpdateScheduleRequest{}),
		Resolve:           adminMutationUpdateScheduleResolver,
		Type:              util.BindOutput(tokenio.UpdateScheduleResponse{}, "Update schedule", nil),
	}
	(*mutation)["pullPayments"] = &graphql.Field{
		DeprecationReason: "use pullTransactions",
		Description:       "Download payments from tokenio",
		Args:              util.BindArguments(tokenio.ProcessPaymentsRequest{}),
		Resolve:           adminMutationPullPaymentsResolver,
		Type:              util.BindOutput(tokenio.ProcessPaymentsResponse{}, "Pull payments", nil),
	}
	(*mutation)["setMatchedPayments"] = &graphql.Field{
		DeprecationReason: "use updateInvoiceItems",
		Description:       "Set payments and status to a schedule",
		Args:              util.BindArguments(tokenio.SetMatchPaymentRequest{}),
		Resolve:           adminMutationSetMatchedPaymentsResolver,
		Type:              util.BindOutput(tokenio.SetMatchPaymentResponse{}, "Update schedule result", nil),
	}
	(*mutation)["createTokenRequest"] = &graphql.Field{
		Description: "Create token request for account information",
		Args:        util.BindArguments(tokenio.GetTokenRequest{}),
		Resolve:     adminMutationCreateTokenResolver,
		Type:        util.BindOutput(tokenio.GetTokenResponse{}, "Create token response", nil),
	}
	(*mutation)["createPayout"] = &graphql.Field{
		Description: "Create payout transfer",
		Args:        util.BindArguments(offer.PayoutRequest{}),
		Resolve:     adminMutationCreatePayoutResolver,
		Type: util.BindOutput(
			offer.PayoutResponse{},
			"Create payout transfer response - return redirect url",
			nil,
		),
	}
}

func translatePayment(p *invoice.Payment) *tokenio.Payment {
	var status tokenio.PaymentStatus

	switch p.GetStatus() {
	case invoice.PaymentStatus_PAYMENT_SUCCESS:
		status = tokenio.PaymentStatus_SUCCESS
	case invoice.PaymentStatus_PAYMENT_PENDING:
		status = tokenio.PaymentStatus_PENDING
	case invoice.PaymentStatus_PAYMENT_FAILED:
		status = tokenio.PaymentStatus_FAILURE
	case invoice.PaymentStatus_PAYMENT_CONFIRM:
		status = tokenio.PaymentStatus_PROCESSING
	case invoice.PaymentStatus_PAYMENT_CANCELED:
		status = tokenio.PaymentStatus_FAILURE
	}

	return &tokenio.Payment{
		Id:           p.GetPaymentId(),
		RefNum:       p.GetSystem().GetSystemReference(),
		Type:         tokenio.PaymentType_SINGLE_PAYMENT,
		Status:       status,
		TranactionId: p.GetSystem().GetSystemReference(),
		TransferId:   p.GetSystem().GetSystemReference(),
		Amount:       p.GetAmount(),
		Description:  p.GetSystem().GetTokenio().GetRequest().GetDescription(),
		CreatedAt:    p.GetCreatedAt(),
		UpdatedAt:    p.GetUpdatedAt(),
		StandingOrderInfo: &tokenio.StandingOrderInfo{
			Frequency: tokenio.StandingOrderFrequency_UNLNOWN,
			StartAt:   p.GetCreatedAt(),
			EndAt:     p.GetCreatedAt(),
		},
		TokenioCreatedAt: p.GetCreatedAt(),
		TransactionType:  tokenio.TransactionType_DEBIT,
		Source:           tokenio.PaymentSource_TRANSFER,
		ReconcileState:   tokenio.PaymentReconcileStatus_RECONCILED,
		Origin:           p.GetSystem().GetTokenio().GetRequest().GetPurpose(),
		Destination:      tokenio.PaymentDestination_PLATFORM,
		Property:         p.GetProperty(),
		PaymentId:        p.GetPaymentId(),
		InvoiceId:        p.GetInvoiceId(),
		CorrelationType:  p.GetCorrelationType(),
		CorrelationId:    p.GetCorrelationId(),
		Payee:            p.GetUser(),
	}
}

func translatePayments(ps []*invoice.Payment) (tokpays []*tokenio.Payment) {
	for _, p := range ps {
		tokpays = append(tokpays, translatePayment(p))
	}

	return
}

func adminQueryGetPaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &tokenio.GetPaymentsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var filters []*q.QueryParamsFilter

	if request.GetStartAt() != nil || request.GetEndAt() != nil {
		start := time.Time{}
		end := time.Now()

		if request.GetStartAt() != nil {
			start = request.GetStartAtValue().AsTime()
		}

		if request.GetEndAt() != nil {
			start = request.GetEndAtValue().AsTime()
		}

		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"created"},
			Value:    []string{start.Format(time.RFC3339), end.Format(time.RFC3339)},
			Operator: q.QueryFilterOperator_RANGE,
		})
	}

	if request.GetAmount() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"amount"},
			Value:    []string{strconv.FormatFloat(request.GetAmountValue(), 'g', -1, 64)},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}

	if request.GetAmountFrom() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"amount"},
			Value:    []string{strconv.FormatFloat(request.GetAmountFromValue(), 'g', -1, 64)},
			Operator: q.QueryFilterOperator_MORE,
		})
	}

	if request.GetAmountTo() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"amount"},
			Value:    []string{strconv.FormatFloat(request.GetAmountToValue(), 'g', -1, 64)},
			Operator: q.QueryFilterOperator_LESS,
		})
	}

	if request.GetProperty() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"property"},
			Value:    []string{request.GetPropertyValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}

	if request.GetRefNumQuery() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"system_reference"},
			Value:    []string{request.GetRefNumQueryValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}

	filters = append(filters, &q.QueryParamsFilter{
		Fields:   []string{"system"},
		Value:    []string{"TOKENIO"},
		Operator: q.QueryFilterOperator_EQUALS,
	})

	payments, err := connections.InvoiceService.PaymentFilterList(params.Context, &invoice.PaymentFilterListRequest{
		Params: &q.QueryParams{
			Pagination: &q.QueryParamsPagination{
				Page:    1,
				PerPage: 100,
			},
			Sort: []*q.QueryParamsSort{
				{
					Field: "created",
					Order: -1,
				},
			},
			Filter: filters,
		},
	})
	if err != nil {
		return nil, err
	}

	return &tokenio.GetPaymentsResponse{
		Payments: translatePayments(payments.GetPayments()),
	}, nil
}

func translateFieldNames(n string) string {
	switch n {
	case "ref_num":
		return "system_reference"
	case "payee":
		return "userid"
	case "created_at":
		return "created"
	case "updated_at":
		return "updated"
	default:
		return n
	}
}

func translateFieldsNames(fields []string) (res []string) {
	for _, n := range fields {
		res = append(res, translateFieldNames(n))
	}

	return
}

func adminQueryGetPaymentsFilterResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	for _, f := range input.Filter {
		f.Fields = translateFieldsNames(f.GetFields())
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created", Order: -1}}
	}

	for _, f := range req.Sort {
		f.Field = translateFieldNames(f.GetField())
	}

	req.Filter = append(req.Filter, &q.QueryParamsFilter{
		Fields:   []string{"system"},
		Value:    []string{"TOKENIO"},
		Operator: q.QueryFilterOperator_EQUALS,
	})

	resp, err := connections.InvoiceService.PaymentFilterList(params.Context, &invoice.PaymentFilterListRequest{
		Params: req,
	})
	if err != nil {
		return nil, err
	}

	return getPaymentsFilterResult{
		Data:  translatePayments(resp.GetPayments()),
		Total: resp.GetTotal(),
	}, nil
}

func translateServiceType(s *extras.ServiceSnapshot, i, max int) (t tokenio.SchedulePaymentType) {
	switch {
	case s.GetType() == extras.ServiceType_DEPOSIT:
		return tokenio.SchedulePaymentType_HOLDING_DEPOSIT
	case s.GetType() == extras.ServiceType_RENT && i == 0:
		return tokenio.SchedulePaymentType_MOVEIN_PAYMENT
	case s.GetType() == extras.ServiceType_RENT && i == max:
		return tokenio.SchedulePaymentType_LAST_PAYMENT
	case s.GetType() == extras.ServiceType_RENT:
		return tokenio.SchedulePaymentType_REGULAR_PAYMENT
	}

	return
}

func translateSchedules(ctx context.Context, ss []*offer.ScheduledPayment) (
	scheduled []*tokenio.SchedulePayment,
	err error,
) {
	for i, s := range ss {
		state := tokenio.ProcessingState_NOT_PROCESSED
		recon := tokenio.PaymentReconcileStatus_NO_MATCH_FOUND

		var amount float64

		var payments []string

		for _, p := range s.GetParts() {
			for _, sc := range p.GetServices() {
				amount += sc.GetPrice() + sc.GetTaxAmount()
			}

			payments = append(payments, p.GetPaymentId())

			if p.GetPaymentStatus() == offer.PaymentStatus_PAYMENT_SUCCESS {
				state = tokenio.ProcessingState_PROCESSED
				recon = tokenio.PaymentReconcileStatus_NO_MATCH_FOUND
			}
		}

		for _, p := range s.GetParts() {
			for _, sc := range p.GetServices() {
				scheduled = append(scheduled, &tokenio.SchedulePayment{
					OrderId:         int32(i),
					Type:            translateServiceType(sc, i, len(ss)-1),
					State:           state,
					ReconcileStatus: recon,
					Amount:          sc.GetPrice(),
					Payments:        payments,
					PayAt:           p.GetPaymentDue(),
					UpdateAt:        sc.GetUpdated(),
				})
			}
		}
	}

	return
}

func adminQueryGetScheduleResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &GetScheduleRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var offerID string

	if request.Offer != "" {
		offerID = request.Offer
	}

	if request.GetCorrelationTypeValue() == correlationTypeOffer {
		offerID = request.GetCorrelationIdValue()
	}

	if offerID == "" {
		return nil, errOfferNotFound
	}

	off, err := connections.OfferService.ServiceGet(params.Context, &offer.GetRequest{
		Id: offerID,
	})
	if err != nil {
		return nil, err
	}

	switch off.GetOffer().GetState() {
	case offer.OfferState_SIGNED, offer.OfferState_AWAITING_CONFIRM_RENT, offer.OfferState_RENTPAID:
	default:
		return &tokenio.GetScheduleResponse{
			Schedule: &tokenio.Schedule{
				Id:              offerID,
				CorrelationType: correlationTypeOffer,
				CorrelationId:   offerID,
				Property:        off.GetOffer().GetProperty(),
				CreatedAt:       off.GetOffer().GetCreatedAtValue(),
				UpdatedAt:       off.GetOffer().GetUpdatedAtValue(),
			},
			Offer:   offerID,
			Success: true,
		}, nil
	}

	schedule, err := connections.OfferService.ScheduledPaymentList(params.Context, &offer.ScheduledPaymentListRequest{
		User:  "",
		Offer: offerID,
		Params: &q.QueryParams{
			Pagination: &q.QueryParamsPagination{
				Page:    1,
				PerPage: 1000,
			},
			Sort: []*q.QueryParamsSort{
				{
					Field: "date",
					Order: 1,
				},
			},
			Filter: nil,
		},
	})
	if err != nil {
		return nil, err
	}

	scheduled, err := translateSchedules(params.Context, schedule.GetItems())
	if err != nil {
		return nil, err
	}

	return &tokenio.GetScheduleResponse{
		Schedule: &tokenio.Schedule{
			Id:              offerID,
			CorrelationType: correlationTypeOffer,
			CorrelationId:   offerID,
			Property:        off.GetOffer().GetProperty(),
			Schedule:        scheduled,
			CreatedAt:       off.GetOffer().GetCreatedAtValue(),
			UpdatedAt:       off.GetOffer().GetUpdatedAtValue(),
		},
		Offer:   offerID,
		Success: true,
	}, nil
}

func adminQuerymatchPaymentResolverFilter(
	schedule *offer.ScheduledPaymentListResponse,
	request *MatchPropertyPaymentRequest,
) []*q.QueryParamsFilter {
	for i, s := range schedule.GetItems() {
		if i == int(request.GetScheduleOrderValue()) {
			for _, p := range s.GetParts() {
				return []*q.QueryParamsFilter{
					{
						Fields:   []string{"invoice_id"},
						Value:    []string{p.GetInvoiceId()},
						Operator: q.QueryFilterOperator_EQUALS,
					},
				}
			}
		}
	}

	return nil
}

func adminQueryMatchPaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &MatchPropertyPaymentRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Offer != "" {
		request.CorrelationType = &tokenio.MatchPropertyPaymentRequest_CorrelationTypeValue{
			CorrelationTypeValue: correlationTypeOffer,
		}

		request.CorrelationId = &tokenio.MatchPropertyPaymentRequest_CorrelationIdValue{
			CorrelationIdValue: request.Offer,
		}
	}

	var filters []*q.QueryParamsFilter

	if request.GetCorrelationType() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"correlation_type"},
			Value:    []string{request.GetCorrelationTypeValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}

	if request.GetCorrelationId() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"correlation_id"},
			Value:    []string{request.GetCorrelationIdValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}

	if request.GetProperty() != nil {
		filters = append(filters, &q.QueryParamsFilter{
			Fields:   []string{"property"},
			Value:    []string{request.GetPropertyIdValue()},
			Operator: q.QueryFilterOperator_EQUALS,
		})
	}

	filters = append(filters, &q.QueryParamsFilter{
		Fields:   []string{"system"},
		Value:    []string{"TOKENIO"},
		Operator: q.QueryFilterOperator_EQUALS,
	})

	if request.GetCorrelationId() != nil && request.GetCorrelationTypeValue() == correlationTypeOffer &&
		request.GetScheduleOrder() != nil {
		schedule, err := connections.OfferService.ScheduledPaymentList(
			params.Context,
			&offer.ScheduledPaymentListRequest{
				User:  "",
				Offer: request.GetCorrelationIdValue(),
				Params: &q.QueryParams{
					Pagination: &q.QueryParamsPagination{
						Page:    1,
						PerPage: 1000,
					},
					Sort: []*q.QueryParamsSort{
						{
							Field: "date",
							Order: 1,
						},
					},
					Filter: nil,
				},
			},
		)
		if err == nil {
			filters = append(filters, adminQuerymatchPaymentResolverFilter(schedule, request)...)
		}
	}

	payments, err := connections.InvoiceService.PaymentFilterList(params.Context, &invoice.PaymentFilterListRequest{
		Params: &q.QueryParams{
			Pagination: &q.QueryParamsPagination{
				Page:    1,
				PerPage: 100,
			},
			Sort: []*q.QueryParamsSort{
				{
					Field: "created",
					Order: -1,
				},
			},
			Filter: filters,
		},
	})
	if err != nil {
		return nil, err
	}

	return &tokenio.MatchPropertyPaymentResponse{
		Payments: translatePayments(payments.GetPayments()),
	}, nil
}

func adminMutationMatchPaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &MatchPropertyPaymentsAllRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Offer != "" {
		request.CorrelationType = &tokenio.MatchPropertyPaymentsAllRequest_CorrelationTypeValue{
			CorrelationTypeValue: correlationTypeOffer,
		}
		request.CorrelationId = &tokenio.MatchPropertyPaymentsAllRequest_CorrelationIdValue{
			CorrelationIdValue: request.Offer,
		}
	}

	return &tokenio.MatchPropertyPaymentsAllResponse{}, nil
}

func adminMutationPullPaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &tokenio.ProcessPaymentsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return &tokenio.ProcessPaymentsResponse{}, nil
}

func adminMutationUpdatePaymentResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &UpdatePaymentRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return &tokenio.UpdatePaymentResponse{}, nil
}

func adminMutationUpdateScheduleResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &UpdateScheduleRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.Schedule != nil && request.Schedule.Offer != "" {
		request.Schedule.CorrelationType = correlationTypeOffer
		request.Schedule.CorrelationId = request.Schedule.Offer
	}

	return &tokenio.UpdateScheduleResponse{}, nil
}

func adminMutationSetMatchedPaymentsResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &tokenio.SetMatchPaymentRequest{}

	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return &tokenio.SetMatchPaymentResponse{}, nil
}

func adminMutationCreateTokenResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &tokenio.GetTokenRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.TokenioService.GetToken(params.Context, request)
}

func adminMutationCreatePayoutResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	request := &offer.PayoutRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.Admin = user.GetAccount().GetIdValue()

	return connections.OfferService.Payout(params.Context, request)
}
