package admin

import (
	"context"

	"github.com/graphql-go/graphql"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	cardpayment "gitlab.mashroom.com/mashroom/backend/cardpayment/api"
	q "gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
)

type allPropertiesResult struct {
	Data  []*property.Property `json:"data,omitempty"`
	Total int32                `json:"total,omitempty"`
}

type allTenanciesResult struct {
	Data  []*property.Tenancy `json:"data,omitempty"`
	Total int32               `json:"total,omitempty"`
}

type propertyData struct {
	*property.Property
	Documents   []*document.Document   `json:"documents,omitempty"`
	ListingFee  []*cardpayment.Receipt `json:"listing_fee,omitempty"`
	AdminVerify *property.AdminVerify  `json:"admin_verify,omitempty"`
	Viewings    []*property.Viewing    `json:"viewings,omitempty"`
}

func (p propertyData) ProvideProperty() *property.Property {
	return p.Property
}

type publishResponse struct {
	Success  bool                     `json:"success,omitempty"`
	Errors   []property.PropertyError `json:"errors,omitempty"`
	ErrorMsg string                   `json:"error_msg,omitempty"`
	Property *propertyData            `json:"property,omitempty"`
}

type withdrawResponse struct {
	Success  bool                     `json:"success,omitempty"`
	Errors   []property.PropertyError `json:"errors,omitempty"`
	ErrorMsg string                   `json:"error_msg,omitempty"`
}

func initProperty(query, mutation *graphql.Fields) {
	(*query)["Property"] = &graphql.Field{
		Description: "Retrieves specified property",
		Args:        util.BindArguments(property.GetRequest{}),
		Resolve:     adminQueryGetPropertyResolver,
		Type:        util.BindOutput(propertyData{}, "Property information", nil),
	}
	(*query)["allProperties"] = &graphql.Field{
		Description: "Retrieves property list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllPropertiesResolver,
		Type:        util.BindOutput(allPropertiesResult{}, "Property list", nil),
	}
	(*query)["allTenancies"] = &graphql.Field{
		Description: "Retrieves tenancy list",
		Args:        util.BindArguments(allParams{}),
		Resolve:     adminQueryAllTenanciesResolver,
		Type:        util.BindOutput(allTenanciesResult{}, "Tenancy list", nil),
	}

	(*mutation)["publishProperty"] = &graphql.Field{
		Description: "Publish property",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationPublishPropertyResolver,
		Type:        util.BindOutput(publishResponse{}, "Publish property response", nil),
	}
	(*mutation)["unpublishProperty"] = &graphql.Field{
		Description: "Unpublish property",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationUnpublishPropertyResolver,
		Type:        util.BindOutput(publishResponse{}, "Unpublish property response", nil),
	}
	(*mutation)["updateProperty"] = &graphql.Field{
		Description: "Update property",
		Args:        util.BindArguments(propertyData{}),
		Resolve:     adminMutationUpdatePropertyResolver,
		Type:        util.BindOutput(propertyData{}, "Property information", nil),
	}
	(*mutation)["reorderPhotos"] = &graphql.Field{
		Description: "Reorder photos",
		Args:        util.BindArguments(property.ReorderPhotosRequest{}),
		Resolve:     adminMutationReorderPhotosResolver,
		Type:        util.BindOutput(property.ReorderPhotosResponse{}, "Reorder information", nil),
	}
	(*mutation)["withdrawProperty"] = &graphql.Field{
		Description: "Withdraw property",
		Args:        util.BindArguments(getParams{}),
		Resolve:     adminMutationWithdrawPropertyResolver,
		Type:        util.BindOutput(withdrawResponse{}, "Withdraw property response", nil),
	}
	(*mutation)["updateRewardStatus"] = &graphql.Field{
		Description: "Update reward status of the viewing",
		Args:        util.BindArguments(property.NewRewardViewingRequest{}),
		Resolve:     adminMutationUpdateRewardStatusResolver,
		Type:        util.BindOutput(property.NewRewardViewingResponse{}, "Update reward status response", nil),
	}
	(*mutation)["addPropertyNote"] = &graphql.Field{
		Description: "Add an admin note to a property",
		Args:        util.BindArguments(property.AddNoteRequest{}),
		Resolve:     adminMutationAddNotePropertyResolver,
		Type:        util.BindOutput(property.AddNoteResponse{}, "Add property note response", nil),
	}
	(*mutation)["setAdmin"] = &graphql.Field{
		Description: "Add an admin user to a property",
		Args:        util.BindArguments(property.SetAdminRequest{}),
		Resolve:     adminMutationSetAdminPropertyResolver,
		Type:        graphql.Boolean,
	}
	(*mutation)["getAdmin"] = &graphql.Field{
		Description: "Get an admin user from a property",
		Args:        util.BindArguments(property.GetAdminRequest{}),
		Resolve:     adminMutationGetAdminPropertyResolver,
		Type:        util.BindOutput(property.GetAdminResponse{}, "Admin user", nil),
	}
}

func enrichPropertyData(ctx context.Context, prop *property.Property, resp *propertyData) {
	if prop == nil {
		return
	}

	// Get documents for property
	reqDoc := &document.ListRequest{
		Property: prop.GetId().GetValue(),
		User:     prop.GetPeople().GetLandlord().GetValue(),
	}
	if docs, err := connections.DocumentService.List(ctx, reqDoc); err == nil {
		resp.Documents = docs.GetDocuments()
	}

	// Get listing fees for property
	req := &q.QueryParams{
		Filter: []*q.QueryParamsFilter{
			{
				Fields:   []string{"property"},
				Value:    []string{prop.GetId().GetValue()},
				Operator: q.QueryFilterOperator_EQUALS,
			},
			{
				Fields:   []string{"userid"},
				Value:    []string{prop.GetPeople().GetLandlord().GetValue()},
				Operator: q.QueryFilterOperator_EQUALS,
			},
			{
				Fields: []string{"correlation_target"},
				Value: []string{
					"listing-services",
					"reference-check",
					"cart",
				},
				Operator: q.QueryFilterOperator_ONEOF,
			},
		},
		Sort: []*q.QueryParamsSort{{
			Field: "id",
			Order: 1,
		}},
		Pagination: &q.QueryParamsPagination{
			Page:    1,
			PerPage: 1000,
		},
	}

	invresp, err := connections.InvoiceService.InvoiceFilterList(ctx, &invoice.InvoiceFilterListRequest{
		Params: req,
	})
	if err != nil {
		return
	}

	resp.ListingFee = append(resp.ListingFee, resolvers.RenderCardpaymentReceipts(invresp.GetInvoices())...)

	// Get Viewings
	if views, err := connections.PropertyService.ListAllView(ctx, &property.ListViewRequest{
		Property: prop.GetId(),
	}); err == nil {
		resp.Viewings = views.List
	}
}

func adminMutationUpdateRewardStatusResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &property.NewRewardViewingRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	return connections.PropertyService.NewRewardViewing(params.Context, request)
}

func adminQueryGetPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	request := &property.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	prop, err := connections.PropertyService.GetAdminVerify(params.Context, request)
	if err != nil {
		return nil, err
	}
	var resp *propertyData = nil
	if prop.GetProperty() != nil {
		resp = &propertyData{
			Property: prop.GetProperty(),
		}
		if prop.GetAdminVerify() != nil {
			resp.AdminVerify = prop.GetAdminVerify()
		}
	}
	if (prop != nil) && (resp != nil) {
		enrichPropertyData(params.Context, prop.GetProperty(), resp)
	}
	return resp, nil
}

func adminQueryAllPropertiesResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	input := &allParams{}
	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}
	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}
	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "created_at", Order: -1}}
	}
	request := &property.PropertyListFilterRequest{
		Params: req,
	}
	request.AdminSite = true
	resp, err := connections.PropertyService.ListFilter(params.Context, request)
	if err != nil {
		return nil, err
	}
	properties := resp.GetProperties()
	total := resp.GetTotal()
	return allPropertiesResult{
		Data:  properties,
		Total: total,
	}, nil
}

func adminQueryAllTenanciesResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	input := &allParams{}

	if err := util.UnmarshalArguments(params.Args, input); err != nil {
		return nil, err
	}

	req := &q.QueryParams{
		Filter: input.Filter,
		Sort:   input.Sort,
		Pagination: &q.QueryParamsPagination{
			Page:    input.Pagination.Page,
			PerPage: input.Pagination.PerPage,
		},
	}

	if input.Sort == nil {
		req.Sort = []*q.QueryParamsSort{{Field: "movedin", Order: -1}}
	}

	resp, err := connections.PropertyService.TenancyListFilter(params.Context, &property.TenancyListFilterRequest{
		Params: req,
	})
	if err != nil {
		return nil, err
	}

	return allTenanciesResult{
		Data:  resp.GetTenancies(),
		Total: resp.GetTotal(),
	}, nil
}

func adminMutationUpdatePropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &propertyData{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.UpdateAndVerifyRequest{
		User:        user.GetAccount().GetIdValue(),
		Property:    req.Property,
		AdminVerify: req.AdminVerify,
	}
	prop, err := connections.PropertyService.UpdateAndVerify(params.Context, request)
	if err != nil {
		return nil, err
	}
	resp := &propertyData{
		Property:    prop.GetProperty(),
		AdminVerify: prop.GetAdminVerify(),
	}
	if prop != nil {
		enrichPropertyData(params.Context, prop.GetProperty(), resp)
	}
	return resp, nil
}

func makePublishResponse(
	ctx context.Context,
	success bool,
	errors []property.PropertyError,
	errmsg string,
	prop *property.Property,
	verify *property.AdminVerify,
) *publishResponse {
	resp := &publishResponse{
		Success:  success,
		Errors:   errors,
		ErrorMsg: errmsg,
	}

	if prop != nil {
		resp.Property = &propertyData{
			Property: prop,
		}

		if verify != nil {
			resp.Property.AdminVerify = verify
		}

		enrichPropertyData(ctx, prop, resp.Property)
	}

	return resp
}

func adminMutationPublishPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.PublishRequest{
		Id:   req.ID,
		User: user.GetAccount().GetIdValue(),
	}
	r, err := connections.PropertyService.ServicePublish(params.Context, request)
	if err != nil {
		return nil, err
	}

	resp := makePublishResponse(
		params.Context,
		len(r.GetErrors()) == 0,
		r.GetErrors(),
		r.GetErrorMsg(),
		r.GetProperty(),
		r.GetAdminVerify(),
	)

	return resp, nil
}

func adminMutationUnpublishPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.UnpublishRequest{
		Id:        req.ID,
		User:      user.GetAccount().GetIdValue(),
		AdminSite: true,
	}

	r, err := connections.PropertyService.ServiceUnpublish(params.Context, request)
	if err != nil {
		return nil, err
	}

	prop, _ := connections.PropertyService.GetAdminVerify(params.Context, &property.GetRequest{
		Id: req.ID,
	})

	resp := makePublishResponse(
		params.Context,
		len(r.GetErrors()) == 0,
		r.GetErrors(),
		r.GetErrorMsg(),
		prop.GetProperty(),
		prop.GetAdminVerify(),
	)

	return resp, nil
}

func adminMutationReorderPhotosResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &property.ReorderPhotosRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.PropertyService.ReorderPhotos(params.Context, req)
}

func adminMutationWithdrawPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &getParams{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.WithdrawnRequest{
		Id:        req.ID,
		User:      user.GetAccount().GetIdValue(),
		AdminSite: true,
	}
	r, err := connections.PropertyService.ServiceWithdrawn(params.Context, request)
	if err != nil {
		return nil, err
	}
	resp := &withdrawResponse{
		Success:  len(r.GetErrors()) == 0,
		Errors:   r.GetErrors(),
		ErrorMsg: r.GetErrorMsg(),
	}

	return resp, nil
}

func adminMutationAddNotePropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}
	req := &property.AddNoteRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	req.User = user.GetAccount().GetIdValue()
	resp, err := connections.PropertyService.AddNote(params.Context, req)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func adminMutationSetAdminPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &property.SetAdminRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	_, err := connections.PropertyService.SetAdmin(params.Context, req)

	return err == nil, err
}

func adminMutationGetAdminPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := resolvers.CheckPermissions(params.Context, resolvers.AdminRole); err != nil {
		return nil, err
	}

	req := &property.GetAdminRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	return connections.PropertyService.GetAdmin(params.Context, req)
}
