//go:generate go-bindata -pkg resolvers -ignore .*\.go .
//go:generate go fmt ./...
//go:generate go fix ./...

package resolvers

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"io/ioutil"
	"net/http"
	"strings"
	"time"

	"github.com/eientei/wsgraphql/mutcontext"
	"github.com/golang/protobuf/ptypes"
	"github.com/graphql-go/graphql"
	uuid "github.com/satori/go.uuid"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/middleware"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers/errs"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	mixpanel "gitlab.mashroom.com/mashroom/backend/mixpanel/api"
)

type UserInfoForm struct {
	RefreshToken string `json:"refresh_token"`
	Email        string `json:"email"`
	Password     string `json:"password"`
	Phone        string `json:"phone"`
	FirstName    string `json:"first_name"`
	LastName     string `json:"last_name"`
}

type UserSignupForm struct {
	UserInfoForm
	Otp         string `json:"otp"`
	TosAccepted uint64 `json:"tos_accepted"`
}

type TosWordpress struct {
	Modified string
	Content  Content
}

type Content struct {
	Rendered string
}

type TosOutput struct {
	Text    string `json:"text"`
	Version int64  `json:"version"`
}

type OtpRequest struct {
	Phone string `json:"phone"`
}

type OtpResponse struct {
	Success bool `json:"success"`
}

type ResendRequest struct {
	Type invitation.ContactValidationType `json:"type"`
	Key  string                           `json:"key"`
}

type ResendResponse struct {
	Success bool `json:"success"`
}

type VerifyRequest struct {
	Password string `json:"password"`
	Otp      string `json:"otp"`
}

type VerifyResponse struct {
	PasswordOk bool `json:"password_ok"`
	OtpOk      bool `json:"otp_ok"`
}

var AnonymousRole = "<anonymous>"
var UserRole = ""
var AdminRole = "admin"

func initAuth(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "AuthQuery",
		Description: "Auth query root",
		Fields: graphql.Fields{
			"tos": &graphql.Field{
				Description: "Returns terms of service",
				Resolve:     AuthQueryTosResolver,
				Type:        util.BindOutput(TosOutput{}, "Tos text and version", nil),
			},
			"session": &graphql.Field{
				Description: "Returns current session",
				Resolve:     AuthQuerySessionResolver,
				Type:        graphql.String,
			},
		},
	}))

	(*rootQueries)["auth"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	verify := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "AuthVerifyMutation",
		Description: "AuthVerify mutation root",
		Fields: graphql.Fields{
			"email": &graphql.Field{
				Description: "Verify email",
				Args:        util.BindArguments(auth.VerifyEmailRequest{}),
				Resolve:     AuthMutationVerifyEmailResolver,
				Type:        util.BindOutput(auth.VerifyEmailResponse{}, "Verify email status", nil),
			},
			"phone": &graphql.Field{
				Description: "Verify phone",
				Args:        util.BindArguments(auth.VerifyPhoneRequest{}),
				Resolve:     AuthMutationVerifyPhoneResolver,
				Type:        util.BindOutput(auth.VerifyPhoneResponse{}, "Verify phone status", nil),
			},
		},
	}))

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "AuthMutation",
		Description: "Auth mutation root",
		Fields: graphql.Fields{
			"login": &graphql.Field{
				Description: "Performs user login by username and password",
				Args:        util.BindArguments(auth.LoginRequest{}),
				Resolve:     AuthMutationLoginResolver,
				Type: util.BindOutput(
					auth.LoginResponse{},
					"Login status and new access and refresh tokens",
					util.EnrichmentMap{
						"account": &util.Enrichment{
							Output: func(bind util.BindType) graphql.Output {
								return util.BindOutput(&account.Account{}, "", nil)
							},
							Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
								resp, ok := p.Source.(*auth.LoginResponse)
								if !ok || resp.GetAccessToken() == "" {
									return &account.Account{}, nil
								}
								ctx := p.Context.Value("globalctx").(mutcontext.MutableContext)
								authresp, err := connections.AuthService.Auth(p.Context, &auth.AuthRequest{
									AccessToken: resp.GetAccessToken(),
									AdminSite:   isAdminSite(ctx),
								})
								if err != nil {
									log.WithError(err).Error()
								}
								return authresp.GetAccount(), err
							},
						},
					},
				),
			},
			"session": &graphql.Field{
				Description: "Resumes previous session",
				Resolve:     AuthMutationSessionResolver,
				Args: graphql.FieldConfigArgument{
					"session": &graphql.ArgumentConfig{
						Type:        graphql.String,
						Description: "Session id",
					},
				},
				Type: util.BindOutput(ws.TransientUser{}, "Transient user session", nil),
			},
			"refresh": &graphql.Field{
				Description: "Refreshes access token",
				Args:        util.BindArguments(auth.RefreshRequest{}),
				Resolve:     AuthMutationRefreshResolver,
				Type: util.BindOutput(
					auth.RefreshResponse{},
					"Refresh status and new access and refresh tokens",
					nil,
				),
			},
			"logout": &graphql.Field{
				Description: "Logs out of current session",
				Resolve:     AuthMutationLogoutResolver,
				Type:        util.BindOutput(auth.LogoutResponse{}, "Logout status", nil),
			},
			"signup": &graphql.Field{
				Description: "Signs up for user account",
				Args:        util.BindArguments(auth.SignupRequest{}),
				Resolve:     AuthMutationSignupResolver,
				Type: util.BindOutput(
					auth.SignupResponse{},
					"New access and refresh token and signup status",
					util.EnrichmentMap{
						"account": &util.Enrichment{
							Output: func(bind util.BindType) graphql.Output {
								return util.BindOutput(&account.Account{}, "", nil)
							},
							Resolver: func(p graphql.ResolveParams, oldvalue interface{}) (i interface{}, e error) {
								resp, ok := p.Source.(*auth.SignupResponse)
								if !ok || resp.GetAccessToken() == "" {
									return &account.Account{}, nil
								}
								authresp, err := connections.AuthService.Auth(p.Context, &auth.AuthRequest{
									AccessToken: resp.GetAccessToken(),
									AdminSite:   false,
								})
								if err != nil {
									log.WithError(err).Error()
								}
								return authresp.GetAccount(), err
							},
						},
					},
				),
			},
			"update": &graphql.Field{
				Description: "Updates user details",
				Args:        util.BindArguments(auth.UpdateRequest{}),
				Resolve:     AuthMutationUpdateResolver,
				Type: util.BindOutput(
					auth.UpdateResponse{},
					"Update status and new access and refresh tokens",
					nil,
				),
			},
			"reset": &graphql.Field{
				Description: "Password reset",
				Args:        util.BindArguments(auth.ForgotPasswordRequest{}),
				Resolve:     AuthMutationResetPasswordResolver,
				Type:        util.BindOutput(auth.ForgotPasswordResponse{}, "Update password status", nil),
			},
			"resend": &graphql.Field{
				Description: "Resends OTP or email verification link challenge",
				Args:        util.BindArguments(ResendRequest{}),
				Resolve:     AuthMutationResendResolver,
				Type:        util.BindOutput(ResendResponse{}, "Result of resend operation", nil),
			},
			"verify": &graphql.Field{
				Type: verify,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return verify, nil
				},
			},
		},
	}))

	(*rootMutations)["auth"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func TryGetTransientUser(accessToken string) *ws.TransientUser {
	parts := strings.Split(accessToken, ".")
	if len(parts) < 2 {
		return nil
	}

	bs, err := base64.RawStdEncoding.DecodeString(parts[1])
	if err != nil {
		return nil
	}

	var tok struct {
		TransientSession string `json:"transient_session"`
	}

	err = json.Unmarshal(bs, &tok)
	if err != nil {
		return nil
	}

	return &ws.TransientUser{
		Session: tok.TransientSession,
	}
}

func isAdminSite(ctx context.Context) bool {
	if adminSite, ok := ctx.Value("adminSite").(bool); ok {
		return adminSite
	}
	return false
}

func HasRole(user *auth.AuthResponse, required string) bool {
	if user == nil {
		return required == AnonymousRole
	}
	if required == UserRole {
		return true
	}
	for _, role := range user.Roles {
		if role == required {
			return true
		}
	}
	return false
}

func CheckPermissions(queryContext context.Context, roles ...string) error {
	ctx := queryContext.Value("globalctx").(mutcontext.MutableContext)
	user := queryContext.Value(ws.CtxUser).(*auth.AuthResponse)
	adminSite := isAdminSite(ctx)
	if adminSite {
		roles = append(roles, AdminRole)
	}

	any := false
	for _, r := range roles {
		if r == AdminRole && !adminSite {
			continue
		}
		any = any || HasRole(user, r)
	}

	if !any {
		return errs.ErrAuthError
	}

	return nil
}

func AuthQueryTosResolver(params graphql.ResolveParams) (interface{}, error) {
	layout := "2006-01-02T15:04:05"

	ctx := params.Context.Value("globalctx").(mutcontext.MutableContext)

	req, err := http.NewRequest(
		"GET",
		"https://wordpress.dev.env.mashroom.com/wp-json/wp/v2/pages/?slug=terms-and-conditions",
		nil,
	)
	if err != nil {
		return nil, err
	}
	client := http.Client{}
	resp, err := client.Do(req)

	defer func() {
		_ = resp.Body.Close()
	}()

	if err != nil {
		return nil, err
	}
	var tosWordpress []TosWordpress
	body, errRead := ioutil.ReadAll(resp.Body)
	if errRead != nil {
		return nil, errRead
	}
	err = json.Unmarshal(body, &tosWordpress)
	if err != nil {
		return nil, err
	}
	timeParse, _ := time.Parse(layout, tosWordpress[0].Modified)
	unix, _ := ptypes.TimestampProto(timeParse)

	adminSite := isAdminSite(ctx)

	prefix := ""
	if adminSite {
		prefix = "ADMIN: "
	}
	return &TosOutput{
		Text:    prefix + tosWordpress[0].Content.Rendered,
		Version: unix.Seconds,
	}, nil
}

func AuthQuerySessionResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, grpcerr.InvalidState
	}

	return tuser.Session, nil
}

func authLoginSuccess(ctx context.Context, resp *auth.AuthResponse, accessToken string, method mixpanel.AuthMethod) {
	mutctx := ctx.Value("globalctx").(mutcontext.MutableContext)

	tuser, _ := ctx.Value(ws.CtxTransientUser).(*ws.TransientUser)

	if tuser != nil {
		_, _ = connections.InvoiceService.OrderPrefaceTransfer(
			ctx,
			&invoice.OrderPrefaceTransferRequest{
				Userid:   resp.GetAccount().GetIdValue(),
				Clientid: tuser.Session,
			},
		)
	}

	if err := trackLogin(ctx, resp.GetAccount().GetIdValue(), method); err != nil {
		log.WithError(err).Error("unable to track login with mixpanel")
	}

	mutctx.Set(ws.CtxAccessToken, accessToken)
	mutctx.Set(ws.CtxUser, resp)
}

func AuthMutationLoginResolver(params graphql.ResolveParams) (interface{}, error) {
	remote := params.Context.Value(ws.CtxRemote).(string)
	request := &auth.LoginRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	ctx := params.Context.Value("globalctx").(mutcontext.MutableContext)
	adminSite := isAdminSite(ctx)
	request.Remote = &auth.LoginRequest_RemoteValue{RemoteValue: remote}
	request.AdminSite = &auth.LoginRequest_AdminSiteValue{AdminSiteValue: adminSite}

	resp, err := connections.AuthService.Login(params.Context, request)

	if resp != nil && resp.Errors == nil {
		resp.Errors = make([]auth.AuthErrors, 0)
	}

	if err != nil || resp == nil || !resp.Success || len(resp.Errors) > 0 {
		return resp, err
	}

	authResp, err := connections.AuthService.Auth(ctx, &auth.AuthRequest{
		AccessToken: resp.GetAccessToken(),
		AdminSite:   adminSite,
	})
	if err != nil {
		return nil, err
	}

	authLoginSuccess(params.Context, authResp, resp.GetAccessToken(), mixpanel.AuthMethod_EMAIL)

	return resp, nil
}

func AuthMutationSessionResolver(params graphql.ResolveParams) (interface{}, error) {
	v, ok := params.Args["session"]
	if !ok {
		return nil, grpcerr.InvalidData
	}

	session, ok := v.(string)
	if !ok {
		return nil, grpcerr.InvalidData
	}

	if !strings.HasPrefix(session, "transient:") {
		return nil, grpcerr.InvalidData
	}

	ctx := params.Context.Value("globalctx").(mutcontext.MutableContext)

	tuser := &ws.TransientUser{
		Session: session,
	}

	ctx.Set(ws.CtxTransientUser, tuser)

	return tuser, nil
}

func AuthMutationRefreshResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &auth.RefreshRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if accessToken, ok := params.Context.Value(ws.CtxAccessToken).(string); ok {
		request.AccessToken = accessToken
	}
	ctx := params.Context.Value("globalctx").(mutcontext.MutableContext)
	adminSite := isAdminSite(ctx)
	request.AdminSite = adminSite
	resp, err := connections.AuthService.Refresh(params.Context, request)
	if err != nil || resp == nil || !resp.Success || len(resp.Errors) > 0 {
		return resp, err
	}
	authResp, err := connections.AuthService.Auth(ctx, &auth.AuthRequest{
		AccessToken: resp.GetAccessToken(),
		AdminSite:   adminSite,
	})
	if err != nil {
		return nil, err
	}

	ctx.Set(ws.CtxAccessToken, resp.GetAccessToken())
	ctx.Set(ws.CtxUser, authResp)
	return resp, nil
}

func AuthMutationLogoutResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := CheckPermissions(params.Context, UserRole); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &auth.LogoutRequest{
		Session: user.Session,
	}
	if accessToken, ok := params.Context.Value(ws.CtxAccessToken).(string); ok {
		request.AccessToken = accessToken
	}
	ctx := params.Context.Value("globalctx").(mutcontext.MutableContext)
	request.AdminSite = isAdminSite(ctx)
	resp, err := connections.AuthService.Logout(params.Context, request)
	if err != nil || resp == nil || !resp.Success || len(resp.Errors) > 0 {
		return resp, err
	}

	ctx.Set(ws.CtxTransientUser, &ws.TransientUser{
		Session: "transient:" + uuid.NewV4().String(),
	})
	ctx.Set(ws.CtxAccessToken, "")
	ctx.Set(ws.CtxUser, (*auth.AuthResponse)(nil))
	return resp, nil
}

func AuthMutationSignupResolver(params graphql.ResolveParams) (interface{}, error) {
	ctx := params.Context.Value("globalctx").(mutcontext.MutableContext)
	adminSite := isAdminSite(ctx)
	// CheckPermissions doesn't cover this case
	if adminSite {
		return nil, errs.ErrWrongQuery
	}

	request := &auth.SignupRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if ok {
		request.TransientSession = &auth.SignupRequest_TransientSessionValue{TransientSessionValue: tuser.Session}
	}

	request.Account.Admin = &account.Account_AdminValue{AdminValue: false}
	remote := params.Context.Value(ws.CtxRemote).(string)
	request.Remote = &auth.SignupRequest_RemoteValue{RemoteValue: remote}

	resp, err := connections.AuthService.Signup(params.Context, request)
	if err != nil || resp == nil || !resp.Success || len(resp.Errors) > 0 {
		return resp, err
	}
	authResp, err := connections.AuthService.Auth(ctx, &auth.AuthRequest{
		AccessToken: resp.GetAccessToken(),
		AdminSite:   false,
	})
	if err != nil {
		return nil, err
	}

	if tuser != nil {
		_, _ = connections.InvoiceService.OrderPrefaceTransfer(
			params.Context,
			&invoice.OrderPrefaceTransferRequest{
				Userid:   authResp.GetAccount().GetIdValue(),
				Clientid: tuser.Session,
			},
		)
	}

	_ = trackSignUp(params.Context, authResp.GetAccount())

	ctx.Set(ws.CtxAccessToken, resp.GetAccessToken())
	ctx.Set(ws.CtxUser, authResp)
	return resp, nil
}

func AuthMutationUpdateResolver(params graphql.ResolveParams) (v interface{}, err error) {
	if err = CheckPermissions(params.Context, UserRole); err != nil {
		return nil, err
	}

	request := &auth.UpdateRequest{}

	if err = util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	if request.Account == nil {
		request.Account = &account.Account{}
	}

	request.Account.Id = user.GetAccount().GetId()

	if accessToken, ok := params.Context.Value(ws.CtxAccessToken).(string); ok {
		request.AccessToken = accessToken
	}

	resp, err := connections.AuthService.Update(params.Context, request)
	if err != nil || !resp.GetSuccess() || len(resp.GetErrors()) > 0 {
		return resp, err
	}

	ctx := params.Context.Value("globalctx").(mutcontext.MutableContext)
	authResp, err := connections.AuthService.Auth(ctx, &auth.AuthRequest{
		AccessToken: resp.GetAccessToken(),
		AdminSite:   isAdminSite(ctx),
	})

	if err != nil {
		return nil, err
	}

	ctx.Set(ws.CtxAccessToken, resp.GetAccessToken())
	ctx.Set(ws.CtxUser, authResp)

	return resp, nil
}

func AuthMutationResendResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &ResendRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	userid := "newuser"
	if user != nil {
		userid = user.GetAccount().GetIdValue()
	}
	resp, err := connections.InvitationService.CreateContactValidation(
		params.Context,
		&invitation.CreateContactValidationRequest{
			Type:      request.Type,
			AccountId: userid,
			Key:       request.Key,
			Force:     true,
		},
	)
	if err != nil {
		return nil, err
	}
	return &ResendResponse{Success: resp.GetSuccess()}, nil
}

func AuthMutationVerifyEmailResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := CheckPermissions(params.Context, UserRole); err != nil {
		return nil, err
	}
	request := &auth.VerifyEmailRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if accessToken, ok := params.Context.Value(ws.CtxAccessToken).(string); ok {
		request.AccessToken = accessToken
	}
	return connections.AuthService.VerifyEmail(params.Context, request)
}

func AuthMutationVerifyPhoneResolver(params graphql.ResolveParams) (interface{}, error) {
	if err := CheckPermissions(params.Context, UserRole); err != nil {
		return nil, err
	}
	request := &auth.VerifyPhoneRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if accessToken, ok := params.Context.Value(ws.CtxAccessToken).(string); ok {
		request.AccessToken = accessToken
	}
	return connections.AuthService.VerifyPhone(params.Context, request)
}

func AuthMutationResetPasswordResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &auth.ForgotPasswordRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if remote, ok := params.Context.Value(ws.CtxRemote).(string); ok {
		request.Remote = &auth.ForgotPasswordRequest_RemoteValue{RemoteValue: remote}
	}
	return connections.AuthService.ForgotPassword(params.Context, request)
}
