package resolvers

import (
	"github.com/graphql-go/graphql"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/middleware"
	insurance "gitlab.mashroom.com/mashroom/backend/insurance/api"
)

func initInsurance(rootQueries, rootMutations *graphql.Fields) {
	contentsInsuranceQuery := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ContentsInsuranceQuery",
		Description: "ContentsInsurance query root",
		Fields: graphql.Fields{
			"getInsurance": &graphql.Field{
				Description: "Get contents insurance by id",
				Resolve:     InsuranceQueryGetContentsInsuranceResolver,
				Type: util.BindOutput(insurance.GetContentsInsuranceResponse{},
					"Get contents insurance response", nil),
			},
			"getPropertySubType": &graphql.Field{
				Description: "Get property sub type by property type",
				Resolve:     InsuranceQueryGetPropertySubTypeResolver,
				Type: util.BindOutput(insurance.GetPropertySubTypeResponse{},
					"Get property sub type response", nil),
			},
		},
	}))

	contentsInsuranceMutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ContentsInsuranceMutation",
		Description: "ContentsInsurance mutation root",
		Fields: graphql.Fields{
			"createInsurance": &graphql.Field{
				Description: "Create contents insurance",
				Resolve:     InsuranceMutationCreateContentsInsuranceResolver,
				Type: util.BindOutput(insurance.CreateContentsInsuranceResponse{},
					"Create contents insurance response", nil),
			},
			"updateInsurance": &graphql.Field{
				Description: "Update contents insurance",
				Resolve:     InsuranceMutationUpdateContentsInsuranceResolver,
				Type: util.BindOutput(insurance.UpdateContentsInsuranceResponse{},
					"Update contents insurance response", nil),
			},
		},
	}))

	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "InsuranceQuery",
		Description: "Insurance query root",
		Fields: graphql.Fields{
			"contentsInsurance": &graphql.Field{
				Type: contentsInsuranceQuery,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return contentsInsuranceQuery, nil
				},
			},
		},
	}))

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "InsuranceMutation",
		Description: "Insurance mutation root",
		Fields: graphql.Fields{
			"contentsInsurance": &graphql.Field{
				Type: contentsInsuranceMutation,
				Resolve: func(p graphql.ResolveParams) (interface{}, error) {
					return contentsInsuranceMutation, nil
				},
			},
		},
	}))

	(*rootQueries)["insurance"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	(*rootMutations)["insurance"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func InsuranceQueryGetContentsInsuranceResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &insurance.GetContentsInsuranceRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.InsuranceService.GetContentsInsurance(params.Context, request)
}

func InsuranceQueryGetPropertySubTypeResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &insurance.GetPropertySubTypeRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.InsuranceService.GetPropertySubType(params.Context, request)
}

func InsuranceMutationCreateContentsInsuranceResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &insurance.CreateContentsInsuranceRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.InsuranceService.CreateContentsInsurance(params.Context, request)
}

func InsuranceMutationUpdateContentsInsuranceResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &insurance.UpdateContentsInsuranceRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.InsuranceService.UpdateContentsInsurance(params.Context, request)
}
