package resolvers

import (
	"context"
	"fmt"

	"github.com/graphql-go/graphql"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/middleware"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers/errs"
	mixpanel "gitlab.mashroom.com/mashroom/backend/mixpanel/api"
)

type MixpanelOneofEvent struct {
	AddToCart         *mixpanel.Event_AddToCart         `json:"add_to_cart,omitempty"`
	BrowseProductPage *mixpanel.Event_BrowseProductPage `json:"browse_product_page,omitempty"`
}

func initMixpanel(rootQueries, rootMutations *graphql.Fields) {
	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "MixpanelMutation",
		Description: "Mixpanel mutation root",
		Fields: graphql.Fields{
			"track": &graphql.Field{
				Description: "Track one of specified event with Mixpanel.",
				Args:        util.BindArguments(&MixpanelOneofEvent{}),
				Resolve:     resolveMixpanelTrack,
				Type:        graphql.Boolean,
			},
		},
	}))

	(*rootMutations)["mixpanel"] = &graphql.Field{
		Type: mutation,
		Resolve: func(graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func resolveMixpanelTrack(params graphql.ResolveParams) (interface{}, error) {
	ip, _ := params.Context.Value(ws.CtxRemote).(string)
	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)

	input := MixpanelOneofEvent{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return false, err
	}

	track := &mixpanel.TrackRequest{
		UserId: user.GetAccount().GetIdValue(),
		IpAddr: ip,
	}

	switch {
	default:
		return false, fmt.Errorf("%w: no events passed", errs.ErrInvalidGQLInput)
	case input.AddToCart != nil:
		track.Event = &mixpanel.TrackRequest_AddToCart{
			AddToCart: &mixpanel.Event_AddToCart{
				ProductName:  input.AddToCart.GetProductName(),
				ProductPrice: input.AddToCart.GetProductPrice(),
				PageFrom:     input.AddToCart.GetPageFrom(),
			},
		}
	case input.BrowseProductPage != nil:
		track.Event = &mixpanel.TrackRequest_BrowseProductPage{
			BrowseProductPage: &mixpanel.Event_BrowseProductPage{
				ProductName:  input.BrowseProductPage.GetProductName(),
				ProductPrice: input.BrowseProductPage.GetProductPrice(),
			},
		}
	}

	_, err := connections.MixpanelService.TrackEvent(params.Context, track)
	if err != nil {
		return false, err
	}

	return true, nil
}

func authMethod(source account.AccountCreationType) mixpanel.AuthMethod {
	switch {
	default:
		return mixpanel.AuthMethod_METHOD_UNSPECIFIED
	case source == account.AccountCreationType_ACCOUNT_CREATION_TYPE_EMAIL:
		return mixpanel.AuthMethod_EMAIL
	case source == account.AccountCreationType_ACCOUNT_CREATION_TYPE_GOOGLE:
		return mixpanel.AuthMethod_GOOGLE_OAUTH
	}
}

// trackSignUp is helper to track SingUp Mixpanel event
func trackSignUp(ctx context.Context, acc *account.Account) error {
	ip, _ := ctx.Value(ws.CtxRemote).(string)
	_, err := connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: acc.GetIdValue(),
		IpAddr: ip,
		Event: &mixpanel.TrackRequest_Signup{
			Signup: &mixpanel.Event_SignUp{
				PrimaryEmail: acc.GetContactValue().GetEmailPrimaryValue(),
				FirstName:    acc.GetDetailsValue().GetFirstNameValue(),
				LastName:     acc.GetDetailsValue().GetLastNameValue(),
				CreatedAt:    acc.GetCreatedAtValue(),
				SignupMethod: authMethod(acc.GetCreationTypeValue()),
			},
		},
	})

	return err
}

// trackLogin helps to track Login event with mixpanel service.
func trackLogin(ctx context.Context, userID string, method mixpanel.AuthMethod) error {
	ip, _ := ctx.Value(ws.CtxRemote).(string)
	_, err := connections.MixpanelService.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: userID,
		IpAddr: ip,
		Event: &mixpanel.TrackRequest_Login{
			Login: &mixpanel.Event_Login{
				LoginMethod: method,
			},
		},
	})

	return err
}
