package resolvers

import (
	"context"
	"errors"
	"fmt"
	"time"

	"google.golang.org/protobuf/types/known/wrapperspb"

	"github.com/golang/protobuf/ptypes"
	"github.com/golang/protobuf/ptypes/timestamp"
	"github.com/golang/protobuf/ptypes/wrappers"
	"github.com/graphql-go/graphql"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/wrap"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/middleware"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers/errs"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
)

type IntentionRequest struct {
	Tenancy        *wrappers.StringValue     `json:"tenancy"`
	Intention      property.TenancyIntention `json:"intention"`
	Date           *timestamp.Timestamp      `json:"termination_date,omitempty"`
	PenaltyConsent *wrappers.BoolValue       `json:"penalty_consent"`
}

type LandlordIntentionRequest struct {
	Tenancy         *wrappers.StringValue     `json:"tenancy"`
	Intention       property.TenancyIntention `json:"intention"`
	Date            *timestamp.Timestamp      `json:"termination_date,omitempty"`
	Property        *wrappers.StringValue     `json:"property"`
	OfferConditions *offer.OfferConditions    `json:"offer_conditions"`
	ExpirationDate  *timestamp.Timestamp      `json:"expiration_date,omitempty"`
}
type LandlordIntentionResponse struct {
	Success bool   `json:"success"`
	Tenancy string `json:"tenancy"`
	Offer   string `json:"new_offer"`
}

type SessionIDRequest struct {
	SessionID string `json:"session_id"`
}

type GetTenancyRequest struct {
	ID string `json:"id"`
}

type GetActiveTenancyRequest struct {
	PropertyID string `json:"property_id"`
}

type GetTenancyByOfferRequest struct {
	OfferID string `json:"offer_id"`
}

type GetTenancyByPropertyRequest struct {
	PropertyID string `json:"property_id"`
}

type GetTenancyResponse struct {
	Success bool              `json:"success"`
	Tenancy *property.Tenancy `json:"tenancy"`
}

type GetTenancyClaimInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

type CreateTenancyClaimInput struct {
	TenancyID string                                  `json:"tenancy_id,omitempty"`
	Damages   []*property.TenancyClaim_Damage_Details `json:"damages,omitempty"`
}

type UpdateTenancyClaimDamageInput struct {
	ByClaimID   string                        `json:"by_claim_id,omitempty"`
	ByTenancyID string                        `json:"by_tenancy_id,omitempty"`
	Damage      *property.TenancyClaim_Damage `json:"damage,omitempty"`
}

type RemoveTenancyClaimDamageInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
	DamageID    string `json:"damage_id,omitempty"`
}

type RemoveTenancyClaimInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

type TenancyClaimSubmitSettlementInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

type TenancyClaimGetSettlementInput struct {
	ByClaimID   string `json:"by_claim_id,omitempty"`
	ByTenancyID string `json:"by_tenancy_id,omitempty"`
}

func CensoredAccount(acc *account.Account) *account.Account {
	return &account.Account{
		Id:        acc.GetId(),
		CreatedAt: acc.GetCreatedAt(),
		Details:   acc.GetDetails(),
		Contact:   acc.GetContact(),
	}
}

func initProperty(rootQueries, rootMutations *graphql.Fields) {
	query := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "PropertyQuery",
		Description: "Property query root",
		Fields: graphql.Fields{
			"get": &graphql.Field{
				Description: "Retrieves specified property",
				Args:        util.BindArguments(property.GetRequest{}),
				Resolve:     PropertyQueryGetResolver,
				Type:        util.BindOutput(property.GetResponse{}, "Property information", nil),
			},
			"getprivate": &graphql.Field{
				Description: "Retrieves specified property",
				Args:        util.BindArguments(property.GetPrivateRequest{}),
				Resolve:     PropertyQueryGetPrivateResolver,
				Type:        util.BindOutput(property.GetResponse{}, "Property information", nil),
			},
			"getprivatepreface": &graphql.Field{
				Description: "Retrieves specified property for preface user",
				Args:        util.BindArguments(property.GetPrivateRequest{}),
				Resolve:     PropertyQueryGetPrivatePrefaceResolver,
				Type:        util.BindOutput(property.GetResponse{}, "Property information", nil),
			},
			"my": &graphql.Field{
				Description: "Retrieves owned property",
				Args:        util.BindArguments(property.PropertyListRequest{}),
				Resolve:     PropertyQueryMyResolver,
				Type:        util.BindOutput(property.PropertyListResponse{}, "Property list information", nil),
			},
			"list": &graphql.Field{
				Description: "Retrieves owned property",
				Args:        util.BindArguments(property.PropertyListRequest{}),
				Resolve:     PropertyQueryListResolver,
				Type:        util.BindOutput(property.PropertyListResponse{}, "Property list information", nil),
			},
			"search": &graphql.Field{
				Description: "Searches property by provided criteria",
				Args:        util.BindArguments(property.PropertySearchRequest{}),
				Resolve:     PropertyQuerySearchResolver,
				Type: util.BindOutput(
					property.PropertySearchResponse{},
					"Property list information with pagination",
					nil,
				),
			},
			"searchinpolygon": &graphql.Field{
				Description: "Searches property within a geographical polygon",
				Args:        util.BindArguments(property.PropertySearchInPolygonRequest{}),
				Resolve:     PropertyQuerySearchInPolygonResolver,
				Type: util.BindOutput(
					property.PropertySearchResponse{},
					"Property list information with pagination",
					nil,
				),
			},
			"listview": &graphql.Field{
				Description: "Lists view requests for property",
				Args:        util.BindArguments(property.ListViewRequest{}),
				Resolve:     PropertyQueryListViewResolver,
				Type:        util.BindOutput(property.ListViewResponse{}, "Property view listing", nil),
			},
			"viewreserved": &graphql.Field{
				Description: "Reserved view intervals",
				Args:        util.BindArguments(property.ViewReservedRequest{}),
				Resolve:     PropertyQueryViewReservedResolver,
				Type:        util.BindOutput(property.ViewReservedResponse{}, "Reserved intervals", nil),
			},
			"gettenancy": &graphql.Field{
				Description: "Retrieves specified tenancy",
				Args:        util.BindArguments(GetTenancyRequest{}),
				Resolve:     PropertyQueryGetTenancyResolver,
				Type:        util.BindOutput(GetTenancyResponse{}, "Tenancy information", nil),
			},
			"gettenancybyoffer": &graphql.Field{
				Description: "Retrieves specified tenancy by offer",
				Args:        util.BindArguments(GetTenancyByOfferRequest{}),
				Resolve:     PropertyQueryGetTenancyByOfferResolver,
				Type:        util.BindOutput(GetTenancyResponse{}, "Tenancy information", nil),
			},
			"gettenancybyproperty": &graphql.Field{
				Description: "Retrieves list of tenancy by property id",
				Args:        util.BindArguments(GetTenancyByPropertyRequest{}),
				Resolve:     PropertyQueryGetTenancyByPropertyResolver,
				Type:        util.BindOutput(property.GetTenancyResponse{}, "List of Tenancy", nil),
			},
			"getactivetenancy": &graphql.Field{
				Description: "Retrieves active tenancy for property",
				Args:        util.BindArguments(GetActiveTenancyRequest{}),
				Resolve:     PropertyQueryGetActiveTenancyResolver,
				Type:        util.BindOutput(GetTenancyResponse{}, "Tenancy information", nil),
			},
			"tenancyClaimGet": &graphql.Field{
				Description: "Retrieve existent tenancy claim using one of provided IDs.",
				Args:        util.BindArguments(GetTenancyClaimInput{}),
				Resolve:     resolveGetTenancyClaim,
				Type:        util.BindOutput(property.TenancyClaim_GetResponse{}, "Complete tenancy claim data", nil),
			},
			"tenancyClaimGetSettlement": &graphql.Field{
				Description: "Retrieve tenancy claim settlement invoice using one of provided IDs.",
				Args:        util.BindArguments(TenancyClaimGetSettlementInput{}),
				Resolve:     resolveTenancyClaimGetSettlement,
				Type:        util.BindOutput(invoice.Invoice{}, "", nil),
			},
			"getaveragerentprices": &graphql.Field{
				Description: "Get average rent prices by bedrooms",
				Args:        util.BindArguments(property.GetAverageRentPricesRequest{}),
				Resolve:     PropertyQueryGetAverageRentPrices,
				Type: util.BindOutput(
					property.GetAverageRentPricesResponse{},
					"Average rent prices",
					nil,
				),
			},
		},
	}))

	(*rootQueries)["property"] = &graphql.Field{
		Type: query,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return query, nil
		},
	}

	mutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "PropertyMutation",
		Description: "Property mutation root",
		Fields: graphql.Fields{
			"create": &graphql.Field{
				Description: "Creates new property",
				Args:        util.BindArguments(property.CreateRequest{}),
				Resolve:     PropertyMutationCreateResolver,
				Type:        util.BindOutput(property.CreateResponse{}, "Create status", nil),
			},
			"update": &graphql.Field{
				Description: "Updates existing property",
				Args:        util.BindArguments(property.UpdateRequest{}),
				Resolve:     PropertyMutationUpdateResolver,
				Type:        util.BindOutput(property.UpdateResponse{}, "Create status", nil),
			},
			"delete": &graphql.Field{
				Description: "Deletes existing property",
				Args:        util.BindArguments(property.DeleteRequest{}),
				Resolve:     PropertyMutationDeleteResolver,
				Type:        util.BindOutput(property.DeleteResponse{}, "Delete status", nil),
			},
			"restore": &graphql.Field{
				Description: "Restore property",
				Args:        util.BindArguments(property.RestoreRequest{}),
				Resolve:     PropertyMutationRestoreResolver,
				Type:        util.BindOutput(property.RestoreResponse{}, "Restore status", nil),
			},
			"publish": &graphql.Field{
				Description: "Publish listing request",
				Args:        util.BindArguments(property.PublishRequest{}),
				Resolve:     PropertyMutationPublishResolver,
				Type:        util.BindOutput(property.PublishResponse{}, "Create listing status", nil),
			},
			"republish": &graphql.Field{
				Description: "Republish listing request",
				Args:        util.BindArguments(property.RepublishRequest{}),
				Resolve:     PropertyMutationRepublishResolver,
				Type:        util.BindOutput(property.RepublishResponse{}, "Recreate listing status", nil),
			},
			"publishconsent": &graphql.Field{
				Description: "Publish listing consent request",
				Args:        util.BindArguments(property.PublishConsentRequest{}),
				Resolve:     PropertyMutationPublishConsentResolver,
				Type:        util.BindOutput(property.PublishConsentResponse{}, "Consent listing status", nil),
			},
			"unpublish": &graphql.Field{
				Description: "Unpublish listing request",
				Args:        util.BindArguments(property.UnpublishRequest{}),
				Resolve:     PropertyMutationUnpublishResolver,
				Type:        util.BindOutput(property.UnpublishResponse{}, "Drop listing status", nil),
			},
			"claim": &graphql.Field{
				Description: "Claim property",
				Args:        util.BindArguments(property.ClaimRequest{}),
				Resolve:     PropertyMutationClaimResolver,
				Type:        util.BindOutput(property.ClaimResponse{}, "Claim status", nil),
			},
			"commitstate": &graphql.Field{
				Description: "Commit state",
				Args:        util.BindArguments(property.CommitStateRequest{}),
				Resolve:     PropertyMutationCommitstateResolver,
				Type:        util.BindOutput(property.CommitStateResponse{}, "Commit status", nil),
			},
			"requestview": &graphql.Field{
				Description: "Request property view",
				Args:        util.BindArguments(property.RequestViewRequest{}),
				Resolve:     PropertyMutationRequestViewResolver,
				Type:        util.BindOutput(property.RequestViewResponse{}, "Request view status", nil),
			},
			"cancelview": &graphql.Field{
				Description: "Cancel property view",
				Args:        util.BindArguments(property.CancelViewRequest{}),
				Resolve:     PropertyMutationCancelViewResolver,
				Type:        util.BindOutput(property.CancelViewResponse{}, "Cancel view status", nil),
			},
			"acceptview": &graphql.Field{
				Description: "Accept property view",
				Args:        util.BindArguments(property.AcceptViewRequest{}),
				Resolve:     PropertyMutationAcceptViewResolver,
				Type:        util.BindOutput(property.AcceptViewResponse{}, "Accept view status", nil),
			},
			"rejectview": &graphql.Field{
				Description: "Reject property view",
				Args:        util.BindArguments(property.RejectViewRequest{}),
				Resolve:     PropertyMutationRejectViewResolver,
				Type:        util.BindOutput(property.RejectViewResponse{}, "Reject view status", nil),
			},
			"completeview": &graphql.Field{
				Description: "Complete property view",
				Args:        util.BindArguments(property.CompleteViewRequest{}),
				Resolve:     PropertyMutationCompleteViewResolver,
				Type:        util.BindOutput(property.CompleteViewResponse{}, "Complete view status", nil),
			},
			"failview": &graphql.Field{
				Description: "Fail property view",
				Args:        util.BindArguments(property.FailViewRequest{}),
				Resolve:     PropertyMutationFailViewResolver,
				Type:        util.BindOutput(property.FailViewResponse{}, "Fail view status", nil),
			},
			"changebankdetails": &graphql.Field{
				Name:    "Change bank details",
				Args:    util.BindArguments(property.ChangeBankDetailsRequest{}),
				Resolve: PropertyMutationChangeBankDetails,
				Type:    util.BindOutput(property.ChangeBankDetailsResponse{}, "Change bank details status", nil),
			},
			"withdraw": &graphql.Field{
				Description: "Withdraw property",
				Args:        util.BindArguments(property.WithdrawnRequest{}),
				Resolve:     adminMutationWithdrawPropertyResolver,
				Type:        util.BindOutput(property.WithdrawnResponse{}, "Withdraw property response", nil),
			},
			"savereminder": &graphql.Field{
				Description: "Save listing reminder",
				Args:        util.BindArguments(property.SaveListingReminderRequest{}),
				Resolve:     PropertyMutationSaveReminderResolver,
				Type: util.BindOutput(
					property.SaveListingReminderResponse{},
					"Save listing reminder response",
					nil,
				),
			},
			"leavefromlisting": &graphql.Field{
				Description: "Leave from listing",
				Args:        util.BindArguments(property.LeaveFromListingFormRequest{}),
				Resolve:     PropertyMutationLeaveFromListingResolver,
				Type: util.BindOutput(
					property.LeaveFromListingFormResponse{},
					"Leave from listing response",
					nil,
				),
			},
			"setLandlordIntention": &graphql.Field{
				Description: "Set landlord intention",
				Args:        util.BindArguments(LandlordIntentionRequest{}),
				Resolve:     PropertySetLandlordIntentionResolver,
				Type:        util.BindOutput(LandlordIntentionResponse{}, "Set landlord intention response", nil),
			},
			"setTenantsIntention": &graphql.Field{
				Description: "Set tenants intention",
				Args:        util.BindArguments(IntentionRequest{}),
				Resolve:     PropertySetTenantsIntentionResolver,
				Type:        util.BindOutput(property.UpdateTenancyResponse{}, "Set tenants intention response", nil),
			},
			"commitpreface": &graphql.Field{
				Description: "Create from preface",
				Args:        util.BindArguments(SessionIDRequest{}),
				Resolve:     PropertyMutationCreateFromPrefaceResolver,
				Type:        util.BindOutput(property.CreateResponse{}, "Create from preface", nil),
			},
			"tenancyClaimCreate": &graphql.Field{
				Description: "Create tenancy claim for completed tenancy (property supervisors only).",
				Args:        util.BindArguments(CreateTenancyClaimInput{}),
				Resolve:     resolveCreateTenancyClaim,
				Type:        util.BindOutput(property.TenancyClaim_CreateResponse{}, "", nil),
			},
			"tenancyClaimUpdateDamage": &graphql.Field{
				Description: "Update specified tenancy claim damage using one of provided IDs.",
				Args:        util.BindArguments(UpdateTenancyClaimDamageInput{}),
				Resolve:     resolveUpdateTenancyClaimDamage,
				Type:        util.BindOutput(property.TenancyClaim_UpdateDamageResponse{}, "", nil),
			},
			"tenancyClaimRemoveDamage": &graphql.Field{
				Description: "Remove specified claim damage using one of provided IDs (property supervisors only).",
				Args:        util.BindArguments(RemoveTenancyClaimDamageInput{}),
				Resolve:     resolveRemoveTenancyClaimDamage,
				Type:        util.BindOutput(property.TenancyClaim_RemoveDamageResponse{}, "", nil),
			},
			"tenancyClaimRemove": &graphql.Field{
				Description: "Completely remove tenancy claim using one of provided IDs (property supervisors only).",
				Args:        util.BindArguments(RemoveTenancyClaimInput{}),
				Resolve:     resolveRemoveTenancyClaim,
				Type:        util.BindOutput(property.TenancyClaim_RemoveResponse{}, "", nil),
			},
			"tenancyClaimSubmitSettlement": &graphql.Field{
				Description: "Move tenancy claim state into SETTLEMENT" +
					" and create invoice to pay damages using one of provided IDs (tenants only).",
				Args:    util.BindArguments(TenancyClaimSubmitSettlementInput{}),
				Resolve: resolveTenancyClaimSubmitSettlement,
				Type:    util.BindOutput(invoice.Invoice{}, "", nil),
			},
		},
	}))

	(*rootMutations)["property"] = &graphql.Field{
		Type: mutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return mutation, nil
		},
	}
}

func ForeignInvite(user *auth.AuthResponse, prop *property.PropertyPeople, inv *property.InvitationStatus) bool {
	if !ForeignProperty(user, prop) {
		return false
	}

	var foreignID, foreignContact bool

	switch user.GetAccount().GetIdValue() {
	case inv.GetInviterId():
	case inv.GetInviteeId():
	default:
		foreignID = true
	}

	switch {
	case user.GetAccount().GetContactValue().GetEmailPrimaryValue() == inv.GetInviteeEmail():
	case user.GetAccount().GetContactValue().GetPhonePrimaryValue() == inv.GetInviteePhone():
	default:
		foreignContact = true

		for _, p := range user.GetAccount().GetContactValue().GetPhonesSecondaryValue().GetValue() {
			if p.GetNumber() == inv.GetInviteePhone() {
				foreignContact = false
			}
		}
	}

	return foreignID && foreignContact
}

func ForeignProperty(user *auth.AuthResponse, prop *property.PropertyPeople) bool {
	if user == nil || prop == nil {
		return true
	}

	if HasRole(user, AdminRole) {
		return false
	}

	switch user.GetAccount().GetIdValue() {
	case prop.GetLandlord().GetValue():
	case prop.GetSubmitter().GetValue():
	case prop.GetManager().GetValue():
	default:
		return true
	}

	return false
}

func PropertyQueryGetPrefaceResolver(params graphql.ResolveParams, session string) (interface{}, error) {
	resp, err := connections.PropertyService.GetPreface(
		params.Context,
		&property.GetPrefaceRequest{
			By: &property.GetPrefaceRequest_ClientId{ClientId: session},
		},
	)
	if err != nil && !errors.Is(err, grpcerr.NotFound) {
		return nil, err
	}

	if resp == nil {
		resp = &property.GetPrefaceResponse{}
	}

	if resp.GetPreface() == nil {
		resp.Preface = &property.Preface{}
	}

	if resp.GetPreface().GetClientId() == "" {
		resp.GetPreface().ClientId = session
	}

	return &property.GetResponse{
		Success:      true,
		Property:     renderPrefaceProperty(resp.GetPreface()),
		Verification: renderPrefaceVerification(resp.GetVerification()),
	}, nil
}

func PropertyQueryGetResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	request := &property.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if tuser != nil && request.GetId() == tuser.Session {
		return PropertyQueryGetPrefaceResolver(params, tuser.Session)
	}

	if user != nil {
		request.User = user.GetAccount().GetIdValue()
	}

	resp, err := connections.PropertyService.Get(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func PropertyQueryGetPrivateResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.GetPrivateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	request.IsAdmin = HasRole(user, AdminRole)

	resp, err := connections.PropertyService.GetPrivate(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQueryGetPrivatePrefaceResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.GetPrivateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	user, _ := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user != nil {
		return PropertyQueryGetPrefaceResolver(params, user.GetSession())
	}

	tuser, _ := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if tuser != nil {
		return PropertyQueryGetPrefaceResolver(params, tuser.Session)
	}

	return nil, errs.ErrNoUser
}

func PropertyQueryListPrefaceResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	resp, err := connections.PropertyService.GetPreface(
		params.Context,
		&property.GetPrefaceRequest{
			By: &property.GetPrefaceRequest_ClientId{ClientId: tuser.Session},
		},
	)
	if err != nil {
		if errors.Is(err, grpcerr.NotFound) {
			return &property.PropertyListResponse{
				Success:    true,
				Properties: []*property.Property{},
				Total:      0,
			}, nil
		}

		return nil, err
	}

	return &property.PropertyListResponse{
		Success: true,
		Properties: []*property.Property{
			renderPrefaceProperty(resp.GetPreface()),
		},
		Total: 1,
	}, nil
}

func PropertyQueryMyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return PropertyQueryListPrefaceResolver(params)
	}
	request := &property.PropertyListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = wrap.StringValueFromOneOf(user.GetAccount().GetId(), user.GetAccount().GetIdValue())
	resp, err := connections.PropertyService.List(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func PropertyQueryListResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return PropertyQueryListPrefaceResolver(params)
	}

	request := &property.PropertyListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = wrap.StringValueFromOneOf(user.GetAccount().GetId(), user.GetAccount().GetIdValue())

	resp, err := connections.PropertyService.List(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQuerySearchResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.PropertySearchRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.Search(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQuerySearchInPolygonResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.PropertySearchInPolygonRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.SearchInPolygon(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQueryViewReservedResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.ViewReservedRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.ViewReserved(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyQueryGetTenancyResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &GetTenancyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.GetTenancy(params.Context, &property.GetTenancyRequest{
		By: &property.GetTenancyRequest_Id{Id: request.ID},
	})
	if err != nil {
		return nil, err
	}

	res := &GetTenancyResponse{
		Success: true,
	}
	if len(resp.Tenancy) > 0 {
		res.Tenancy = resp.Tenancy[0]
	}
	return res, nil
}

func PropertyQueryGetTenancyByOfferResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &GetTenancyByOfferRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.GetTenancy(params.Context, &property.GetTenancyRequest{
		By: &property.GetTenancyRequest_Offer{Offer: request.OfferID},
	})
	if err != nil {
		return nil, err
	}

	res := &GetTenancyResponse{
		Success: true,
	}
	if len(resp.Tenancy) > 0 {
		res.Tenancy = resp.Tenancy[0]
	}
	return res, nil
}

func PropertyQueryGetTenancyByPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	request := &GetTenancyByPropertyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.GetTenancyByUser(params.Context, &property.GetTenancyRequest{
		By:   &property.GetTenancyRequest_Property{Property: request.PropertyID},
		User: user.GetAccount().GetIdValue(),
	})

	return resp, err
}

func PropertyQueryGetActiveTenancyResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &GetActiveTenancyRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.GetActiveTenancy(params.Context, &property.GetActiveTenancyRequest{
		Property: request.PropertyID,
	})
	if err != nil {
		return nil, err
	}

	return &GetTenancyResponse{
		Success: true,
		Tenancy: resp.Tenancy,
	}, nil
}

func resolveGetTenancyClaim(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := GetTenancyClaimInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &property.TenancyClaim_GetRequest{
		Sender: user.GetAccount().GetIdValue(),
	}
	switch {
	case input.ByClaimID != "":
		forward.ClaimBy = &property.TenancyClaim_GetRequest_Id{Id: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.ClaimBy = &property.TenancyClaim_GetRequest_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	return connections.PropertyService.TenancyClaimGet(params.Context, forward)
}

func resolveTenancyClaimGetSettlement(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := TenancyClaimGetSettlementInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &property.TenancyClaim_SettlementRequest{
		Sender: user.GetAccount().GetIdValue(),
	}
	switch {
	case input.ByClaimID != "":
		forward.ClaimBy = &property.TenancyClaim_SettlementRequest_Id{Id: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.ClaimBy = &property.TenancyClaim_SettlementRequest_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	settlement, err := connections.PropertyService.TenancyClaimGetSettlement(params.Context, forward)
	if err != nil {
		return nil, err
	}

	// TODO resolve user matching
	// Any from tenants can request settlement after creation,
	// but invoice is associated with only one creator (tenant).
	// Thus here we can potentially get an error, if different tenant user make source GQL-request.
	// This case can be resolved by adding creator ID to settlement data.

	inv, err := connections.InvoiceService.InvoiceGet(params.Context, &invoice.InvoiceGetRequest{
		InvoiceId:        settlement.GetSettlement().GetInvoiceId(),
		InvoiceReference: settlement.GetSettlement().GetInvoiceReference(),
		User:             user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return inv.GetInvoice(), nil
}

func renderPrefaceProperty(p *property.Preface) *property.Property {
	return &property.Property{
		Id: &wrappers.StringValue{
			Value: p.GetClientId(),
		},
		CreatedAt:   p.GetCreatedAt(),
		UpdatedAt:   p.GetUpdatedAt(),
		Location:    p.GetLocation(),
		Details:     p.GetDetails(),
		Rental:      p.GetRental(),
		Description: p.GetDescription(),
	}
}

func renderPrefaceVerification(p *property.PrefaceVerification) *property.PropertyVerification {
	return &property.PropertyVerification{
		Location:    p.GetLocation(),
		Details:     p.GetDetails(),
		Rental:      p.GetRental(),
		Description: p.GetDescription(),
	}
}

func PropertyQueryGetAverageRentPrices(params graphql.ResolveParams) (interface{}, error) {
	request := &property.GetAverageRentPricesRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.GetAverageRentPrices(params.Context, request)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func PropertyMutationCreateTransientResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	request := &property.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.SavePreface(
		params.Context,
		&property.SavePrefaceRequest{
			ClientId:    tuser.Session,
			Location:    request.GetProperty().GetLocation(),
			Details:     request.GetProperty().GetDetails(),
			Rental:      request.GetProperty().GetRental(),
			Description: request.GetProperty().GetDescription(),
		},
	)
	if err != nil {
		return nil, err
	}

	return &property.CreateResponse{
		Success:      true,
		Property:     renderPrefaceProperty(resp.GetPreface()),
		Verification: renderPrefaceVerification(resp.GetVerification()),
	}, nil
}

func PropertyMutationCreateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return PropertyMutationCreateTransientResolver(params)
	}

	request := &property.CreateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.PropertyService.Create(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func PropertyMutationUpdateTransientResolver(params graphql.ResolveParams) (interface{}, error) {
	tuser, ok := params.Context.Value(ws.CtxTransientUser).(*ws.TransientUser)
	if !ok {
		return nil, fmt.Errorf("%w: empty clientID", errs.ErrInvalidGQLInput)
	}

	request := &property.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.SavePreface(
		params.Context,
		&property.SavePrefaceRequest{
			ClientId:    tuser.Session,
			Location:    request.GetProperty().GetLocation(),
			Details:     request.GetProperty().GetDetails(),
			Rental:      request.GetProperty().GetRental(),
			Description: request.GetProperty().GetDescription(),
		},
	)
	if err != nil {
		return nil, err
	}

	return &property.UpdateResponse{
		Success:      true,
		Property:     renderPrefaceProperty(resp.GetPreface()),
		Verification: renderPrefaceVerification(resp.GetVerification()),
	}, nil
}

func PropertyMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return PropertyMutationUpdateTransientResolver(params)
	}

	request := &property.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.PropertyService.Update(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, nil
}

func PropertyMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.DeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Delete(params.Context, request)
}

func PropertyMutationRestoreResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.RestoreRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Restore(params.Context, request)
}

func PropertyMutationPublishResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.PublishRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Publish(params.Context, request)
}

func PropertyMutationRepublishResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.RepublishRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Republish(params.Context, request)
}

func PropertyMutationPublishConsentResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.PublishConsentRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.PublishConsent(params.Context, request)
}

func PropertyMutationUnpublishResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.UnpublishRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Unpublish(params.Context, request)
}

func PropertyMutationClaimResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.ClaimRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.Claim(params.Context, request)
}

func PropertyMutationCommitstateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.CommitStateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.CommitState(params.Context, request)
}

func PropertyQueryListViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.ListViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	resp, err := connections.PropertyService.ListView(params.Context, request)
	if err != nil {
		return nil, err
	}
	return resp, err
}

func PropertyMutationRequestViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.RequestViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.PropertyService.RequestView(context.Background(), request)
}

func PropertyMutationCancelViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.CancelViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.CancelView(context.Background(), request)
}

func PropertyMutationAcceptViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &property.AcceptViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	request.User = user.GetAccount().GetIdValue()

	return connections.PropertyService.AcceptView(context.Background(), request)
}

func PropertyMutationRejectViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.RejectViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.RejectView(context.Background(), request)
}

func PropertyMutationCompleteViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.CompleteViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.CompleteView(context.Background(), request)
}

func PropertyMutationFailViewResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.FailViewRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.FailView(context.Background(), request)
}

func PropertyMutationChangeBankDetails(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.ChangeBankDetailsRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	if remote, ok := params.Context.Value(ws.CtxRemote).(string); ok {
		request.Remote = remote
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.ChangeBankDetails(context.Background(), request)
}

func adminMutationWithdrawPropertyResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	req := &property.WithdrawnRequest{}
	if err := util.UnmarshalArguments(params.Args, req); err != nil {
		return nil, err
	}

	request := &property.WithdrawnRequest{
		Id:        req.Id,
		User:      user.GetAccount().GetIdValue(),
		AdminSite: false,
	}

	return connections.PropertyService.ServiceWithdrawn(params.Context, request)
}

func PropertyMutationSaveReminderResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &property.SaveListingReminderRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.PropertyService.SaveListingReminder(params.Context, request)
}

func PropertyMutationLeaveFromListingResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	request := &property.LeaveFromListingFormRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	request.User = user.GetAccount().GetIdValue()
	return connections.PropertyService.LeaveFromListingForm(params.Context, request)
}

func PropertySetLandlordIntentionResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &LandlordIntentionRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	ctx := params.Context

	var tenancyID string

	var tenancy *property.Tenancy

	if request.Tenancy != nil {
		t, err := connections.PropertyService.GetTenancy(ctx, &property.GetTenancyRequest{
			By: &property.GetTenancyRequest_Id{
				Id: request.Tenancy.Value,
			}})
		if err != nil {
			return nil, err
		}

		if len(t.Tenancy) > 0 {
			tenancyID = t.Tenancy[0].Id
			tenancy = t.Tenancy[0]
		}
	} else {
		t, err := connections.PropertyService.GetActiveTenancy(ctx, &property.GetActiveTenancyRequest{
			Property: request.Property.Value,
		})
		if err != nil {
			return nil, err
		}

		if t.Tenancy != nil {
			tenancyID = t.Tenancy.Id
			tenancy = t.Tenancy
		}
	}

	if tenancy == nil {
		return nil, grpcerr.New(grpcerr.InvalidData, "tenancy not found")
	}

	if tenancy.GetOffer().GetValue() == "" {
		return nil, grpcerr.New(grpcerr.InvalidData, "tenancy has an empty offer")
	}

	offr, err := connections.OfferService.ServiceGet(ctx, &offer.GetRequest{Id: tenancy.Offer.Value})
	if err != nil {
		return nil, err
	}

	if offr.GetOffer().GetLandlord().GetUser() != user.GetAccount().GetIdValue() {
		return nil, grpcerr.New(grpcerr.AuthError, "user is not landlord")
	}

	start := offr.GetOffer().GetConditions().GetMoveinDate().AsTime()
	end := start.AddDate(0, int(offr.GetOffer().GetConditions().GetTermMonths().GetValue()), 0)
	now := time.Now()

	if end.Before(now) {
		return nil, grpcerr.New(grpcerr.InvalidData, "contract is already terminated")
	}

	landlordIntentionDate, _ := ptypes.TimestampProto(now)

	// update tenancy
	_, err = connections.PropertyService.UpdateTenancy(ctx, &property.UpdateTenancyRequest{
		Tenancy: &property.TenancyRequest{
			Id: wrap.StringValue(&tenancyID),
			LandlordIntention: &property.IntentionRequest{
				Intention: request.Intention,
			},
			LandlordIntentionDate: landlordIntentionDate,
			ExpirationDate:        request.ExpirationDate,
			TerminationDate:       request.Date,
		},
	})
	if err != nil {
		return nil, err
	}

	if request.Intention == property.TenancyIntention_INTENTION_RENEW {
		return renewTenancyLandlord(ctx, request, tenancy, user)
	}

	return &LandlordIntentionResponse{
		Success: true,
		Tenancy: tenancyID,
	}, nil
}

func PropertySetTenantsIntentionResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &IntentionRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	var tenancy *property.Tenancy

	t, err := connections.PropertyService.GetTenancy(params.Context, &property.GetTenancyRequest{
		By: &property.GetTenancyRequest_Id{
			Id: request.Tenancy.GetValue(),
		}})
	if err != nil {
		return nil, err
	}

	if len(t.Tenancy) > 0 {
		tenancy = t.Tenancy[0]
	}

	if tenancy == nil {
		return nil, grpcerr.New(grpcerr.InvalidData, "tenancy not found")
	}

	if tenancy.GetOffer().GetValue() == "" {
		return nil, grpcerr.New(grpcerr.InvalidData, "tenancy has an empty offer")
	}

	offr, err := connections.OfferService.ServiceGet(params.Context, &offer.GetRequest{Id: tenancy.Offer.Value})
	if err != nil {
		return nil, err
	}

	var found bool

	for _, tn := range offr.GetOffer().GetTenants().GetValue() {
		if tn.GetUser() == user.GetAccount().GetIdValue() {
			found = true
			break
		}
	}

	if !found {
		return nil, grpcerr.New(grpcerr.AuthError, "user is not a tenant")
	}

	start := offr.GetOffer().GetConditions().GetMoveinDate().AsTime()
	end := start.AddDate(0, int(offr.GetOffer().GetConditions().GetTermMonths().GetValue()), 0)
	now := time.Now()

	if end.Before(now) {
		return nil, grpcerr.New(grpcerr.InvalidData, "contract is already terminated")
	}

	var penaltyConsentDate *timestamp.Timestamp

	if request.PenaltyConsent != nil {
		penaltyConsentDate, _ = ptypes.TimestampProto(now)
	}

	tenantsIntentionDate, _ := ptypes.TimestampProto(now)

	return connections.PropertyService.UpdateTenancy(params.Context, &property.UpdateTenancyRequest{
		Tenancy: &property.TenancyRequest{
			Id: request.Tenancy,
			TenantsIntention: &property.IntentionRequest{
				Intention: request.Intention,
			},
			TenantsIntentionDate: tenantsIntentionDate,
			TerminationDate:      request.Date,
			PenaltyConsentDate:   penaltyConsentDate,
		},
	})
}

func PropertyMutationCreateFromPrefaceResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}

	request := &SessionIDRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.CreateFromPreface(params.Context, &property.CreateFromPrefaceRequest{
		By: &property.CreateFromPrefaceRequest_ClientId{
			ClientId: request.SessionID,
		},
		Owner: user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func resolveCreateTenancyClaim(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := CreateTenancyClaimInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	resp, err := connections.PropertyService.TenancyClaimCreate(params.Context, &property.TenancyClaim_CreateRequest{
		Sender:    user.GetAccount().GetIdValue(),
		TenancyId: input.TenancyID,
		Details:   input.Damages,
	})
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func resolveUpdateTenancyClaimDamage(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := UpdateTenancyClaimDamageInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &property.TenancyClaim_UpdateDamageRequest{
		Sender: user.GetAccount().GetIdValue(),
		Damage: input.Damage,
	}
	switch {
	case input.ByClaimID != "":
		forward.ClaimBy = &property.TenancyClaim_UpdateDamageRequest_Id{Id: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.ClaimBy = &property.TenancyClaim_UpdateDamageRequest_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	return connections.PropertyService.TenancyClaimUpdateDamage(params.Context, forward)
}

func resolveRemoveTenancyClaimDamage(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := RemoveTenancyClaimDamageInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &property.TenancyClaim_RemoveDamageRequest{
		Sender:   user.GetAccount().GetIdValue(),
		DamageId: input.DamageID,
	}
	switch {
	case input.ByClaimID != "":
		forward.ClaimBy = &property.TenancyClaim_RemoveDamageRequest_Id{Id: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.ClaimBy = &property.TenancyClaim_RemoveDamageRequest_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	return connections.PropertyService.TenancyClaimRemoveDamage(params.Context, forward)
}

func resolveRemoveTenancyClaim(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := RemoveTenancyClaimInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &property.TenancyClaim_RemoveRequest{
		Sender: user.GetAccount().GetIdValue(),
	}
	switch {
	case input.ByClaimID != "":
		forward.ClaimBy = &property.TenancyClaim_RemoveRequest_Id{Id: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.ClaimBy = &property.TenancyClaim_RemoveRequest_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	return connections.PropertyService.TenancyClaimRemove(params.Context, forward)
}

func resolveTenancyClaimSubmitSettlement(params graphql.ResolveParams) (interface{}, error) {
	user, ok := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if !ok || user == nil {
		return nil, errs.ErrNoUser
	}

	input := TenancyClaimSubmitSettlementInput{}
	if err := util.UnmarshalArguments(params.Args, &input); err != nil {
		return nil, err
	}

	forward := &property.TenancyClaim_SettlementRequest{
		Sender: user.GetAccount().GetIdValue(),
	}
	switch {
	case input.ByClaimID != "":
		forward.ClaimBy = &property.TenancyClaim_SettlementRequest_Id{Id: input.ByClaimID}
	case input.ByTenancyID != "":
		forward.ClaimBy = &property.TenancyClaim_SettlementRequest_TenancyId{TenancyId: input.ByTenancyID}
	default:
		return nil, fmt.Errorf("%w: neither `by_claim_id' nor 'by_tenancy_id' is specified", errs.ErrInvalidGQLInput)
	}

	settlement, err := connections.PropertyService.TenancyClaimSubmitSettlement(params.Context, forward)
	if err != nil {
		return nil, err
	}

	inv, err := connections.InvoiceService.InvoiceGet(params.Context, &invoice.InvoiceGetRequest{
		InvoiceId:        settlement.GetSettlement().GetInvoiceId(),
		InvoiceReference: settlement.GetSettlement().GetInvoiceReference(),
		User:             user.GetAccount().GetIdValue(),
	})
	if err != nil {
		return nil, err
	}

	return inv.GetInvoice(), nil
}

func renewTenancyLandlord(
	ctx context.Context,
	req *LandlordIntentionRequest,
	tenancy *property.Tenancy,
	user *auth.AuthResponse) (*LandlordIntentionResponse, error) {
	var offerID string

	var updateTenancy bool

	if tenancy.GetNewOffer().GetValue() != "" {
		ex, _ := connections.OfferService.ServiceGet(ctx, &offer.GetRequest{
			Id: tenancy.GetNewOffer().GetValue(),
		})

		if ex.GetOffer().GetState() == offer.OfferState_DENIED {
			tenancy.NewOffer = nil
			updateTenancy = true
		}
	}

	if tenancy.GetNewOffer().GetValue() != "" {
		resp, err := connections.OfferService.Update(ctx, &offer.UpdateRequest{
			User: user.GetAccount().GetIdValue(),
			Offer: &offer.Offer{
				Id:         tenancy.GetNewOffer().GetValue(),
				Conditions: req.OfferConditions,
			},
		})
		if err != nil {
			return nil, err
		}

		offerID = resp.GetOffer().GetId()
	} else {
		resp, err := connections.OfferService.CreateForRenew(ctx, &offer.CreateForRenewRequest{
			User:    user.GetAccount().GetIdValue(),
			Offer:   &offer.Offer{Conditions: req.OfferConditions},
			Tenancy: tenancy.GetId(),
		})
		if err != nil {
			return nil, err
		}

		offerID = resp.GetOffer().GetId()
		updateTenancy = true
	}

	if updateTenancy {
		_, err := connections.PropertyService.UpdateTenancy(ctx, &property.UpdateTenancyRequest{
			Tenancy: &property.TenancyRequest{
				Id:       wrapperspb.String(tenancy.GetId()),
				NewOffer: wrapperspb.String(offerID),
			},
		})
		if err != nil {
			return nil, err
		}
	}

	return &LandlordIntentionResponse{
		Success: true,
		Tenancy: tenancy.Id,
		Offer:   offerID,
	}, nil
}
