package resolvers

import (
	"github.com/graphql-go/graphql"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/common/graphql/util"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	extras "gitlab.mashroom.com/mashroom/backend/extras/api"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/middleware"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers/errs"
)

func initService(rootQueries, rootMutations *graphql.Fields) {
	createQuery := func(title, name, depr string) *graphql.Object {
		getArg := util.BindArguments(extras.GetRequest{})
		listArg := util.BindArguments(extras.ListRequest{})

		if depr == "" {
			delete(getArg, "property_id")
			delete(listArg, "property_id")
		}

		fields := graphql.Fields{
			"get": &graphql.Field{
				DeprecationReason: depr,
				Description:       "Get " + name + " by id",
				Args:              getArg,
				Resolve:           ServicesQueryGetResolver,
				Type:              util.BindOutput(extras.GetResponse{}, title+" description", nil),
			},
			"list": &graphql.Field{
				DeprecationReason: depr,
				Description:       "Get " + name + " list",
				Args:              listArg,
				Resolve:           ServicesQueryListResolver,
				Type:              util.BindOutput(extras.ListResponse{}, title+" list", nil),
			},
		}

		if depr != "" {
			fields["listwithprice"] = &graphql.Field{
				DeprecationReason: depr + "Use get / list",
				Description:       "Get " + name + " list",
				Args:              util.BindArguments(extras.ListRequest{}),
				Resolve:           ServicesQueryListWithPriceResolver,
				Type:              util.BindOutput(extras.ListResponse{}, title+" list with calculated prices", nil),
			}
		}

		return graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
			Name:        title + "Query",
			Description: title + " query root",
			Fields:      fields,
		}))
	}

	servicesQuery := createQuery("Services", "service", "Use products; ")

	(*rootQueries)["services"] = &graphql.Field{
		Type: servicesQuery,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return servicesQuery, nil
		},
	}

	productsQuery := createQuery("Products", "product", "")

	(*rootQueries)["products"] = &graphql.Field{
		Type: productsQuery,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return productsQuery, nil
		},
	}

	servicesMutation := graphql.NewObject(middleware.GraphqlMiddlewareObject(graphql.ObjectConfig{
		Name:        "ServicesMutation",
		Description: "Services mutation root",
		Fields: graphql.Fields{
			"add": &graphql.Field{
				DeprecationReason: "use admin Product mutation",
				Description:       "Add service",
				Args:              util.BindArguments(extras.AddRequest{}),
				Resolve:           ServicesMutationAddResolver,
				Type:              util.BindOutput(extras.AddResponse{}, "Services add", nil),
			},
			"update": &graphql.Field{
				DeprecationReason: "use admin Product mutation",
				Description:       "Update service",
				Args:              util.BindArguments(extras.UpdateRequest{}),
				Resolve:           ServicesMutationUpdateResolver,
				Type:              util.BindOutput(extras.UpdateResponse{}, "Services update", nil),
			},
			"delete": &graphql.Field{
				DeprecationReason: "use admin Product mutation",
				Description:       "Delete service",
				Args:              util.BindArguments(extras.DeleteRequest{}),
				Resolve:           ServicesMutationDeleteResolver,
				Type:              util.BindOutput(extras.DeleteResponse{}, "Services delete", nil),
			},
		},
	}))

	(*rootMutations)["services"] = &graphql.Field{
		Type: servicesMutation,
		Resolve: func(p graphql.ResolveParams) (interface{}, error) {
			return servicesMutation, nil
		},
	}
}

func ServicesQueryGetResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &extras.GetRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.GetProperty() == nil {
		request.Property = &extras.Property{}
	}

	if request.GetProperty().GetId() == "" {
		// nolint:staticcheck // request.GetPropertyId is DEPRECATED
		request.Property.Id = request.GetPropertyId()
	}

	return connections.ExtrasService.Get(params.Context, request)
}

func ServicesQueryListResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &extras.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.GetProperty() == nil {
		request.Property = &extras.Property{}
	}

	if request.GetProperty().GetId() == "" {
		// nolint:staticcheck // request.GetPropertyId is DEPRECATED
		request.Property.Id = request.GetPropertyId()
	}

	return connections.ExtrasService.List(params.Context, request)
}

func ServicesQueryListWithPriceResolver(params graphql.ResolveParams) (interface{}, error) {
	request := &extras.ListRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}

	if request.GetProperty() == nil {
		request.Property = &extras.Property{}
	}

	if request.GetProperty().GetId() == "" {
		// nolint:staticcheck // request.GetPropertyId is DEPRECATED
		request.Property.Id = request.GetPropertyId()
	}

	return connections.ExtrasService.ListWithPrice(params.Context, request)
}

func ServicesMutationAddResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	if !HasRole(user, ContentManagerRole) {
		return nil, errs.ErrAuthError
	}

	request := &extras.AddRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.ExtrasService.Add(params.Context, request)
}

func ServicesMutationUpdateResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	if !HasRole(user, ContentManagerRole) {
		return nil, errs.ErrAuthError
	}

	request := &extras.UpdateRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.ExtrasService.Update(params.Context, request)
}

func ServicesMutationDeleteResolver(params graphql.ResolveParams) (interface{}, error) {
	user := params.Context.Value(ws.CtxUser).(*auth.AuthResponse)
	if user == nil {
		return nil, errs.ErrNoUser
	}
	if !HasRole(user, ContentManagerRole) {
		return nil, errs.ErrAuthError
	}

	request := &extras.DeleteRequest{}
	if err := util.UnmarshalArguments(params.Args, request); err != nil {
		return nil, err
	}
	return connections.ExtrasService.Delete(params.Context, request)
}
