package rest

import (
	"errors"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"

	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
	tokenio "gitlab.mashroom.com/mashroom/backend/tokenio/api"
)

const (
	errorURL      = "https://%s/dashboard"
	dashboardURL  = "https://%s/dashboard"
	adminURL      = "https://admin.%s"
	termsDeclined = "terms declined"
)

func onError(w http.ResponseWriter, err error) {
	log.WithError(err).Error("TokenioHandler")
	w.Header().Set("Location", fmt.Sprintf(errorURL, connections.SiteURL))
	w.WriteHeader(http.StatusMovedPermanently)
	_, _ = w.Write(([]byte)(err.Error()))
}

func TokenioPaymentHandler(w http.ResponseWriter, r *http.Request) {
	resp, err := connections.TokenioService.PaymentHandler(r.Context(), &tokenio.PaymentHandlerRequest{
		Data: []byte(r.URL.RawQuery),
	})
	if err != nil {
		log.WithError(err).Info("TokenioPaymentHandler TokenioService.PaymentHandler\n")
		onError(w, err)
		return
	}

	inv, err := connections.InvoiceService.InvoiceGet(r.Context(), &invoice.InvoiceGetRequest{
		InvoiceId: resp.GetInvoiceId(),
	})
	if err != nil {
		log.WithError(err).Infof("TokenioPaymentHandler InvoiceService.InvoiceGet invoiceID=%+v\n", resp.GetInvoiceId())
		onError(w, err)
		return
	}

	errorStr := r.URL.Query().Get("error_description")

	// check decline error
	if errorStr != "" && strings.ToLower(errorStr) == termsDeclined {
		if inv.GetInvoice().GetCorrelationType() != "offer" {
			onError(w, errors.New("payment declined"))
			return
		}

		offr, err := connections.OfferService.ServiceGet(r.Context(), &offer.GetRequest{
			Id: inv.GetInvoice().GetCorrelationId(),
		})
		backLink := fmt.Sprintf(dashboardURL, connections.SiteURL)
		if err != nil {
			log.WithError(err).
				Infof(
					"TokenioPaymentHandler OfferService.ServiceGet tResp.Offer=%+v\n",
					inv.GetInvoice().GetCorrelationId(),
				)
		} else {
			if offr.GetOffer().State == offer.OfferState_SIGNED {
				backLink = fmt.Sprintf(
					dashboardURL+"/rent?offer=%s",
					connections.SiteURL,
					inv.GetInvoice().GetCorrelationId(),
				)
			} else {
				backLink = fmt.Sprintf(
					dashboardURL+"/deposit?offer=%s",
					connections.SiteURL,
					inv.GetInvoice().GetCorrelationId(),
				)
			}
		}

		w.Header().Set("Location", backLink)
		w.WriteHeader(http.StatusFound)

		return
	}

	w.Header().Set(
		"Location",
		fmt.Sprintf(
			dashboardURL+"/%s/overview",
			connections.SiteURL,
			inv.GetInvoice().GetProperty(),
		),
	)
	w.WriteHeader(http.StatusFound)
}

func TokenioTokeninfoHandler(w http.ResponseWriter, r *http.Request) {
	body, err := ioutil.ReadAll(r.Body)

	if len(body) == 0 && r.URL != nil {
		body = []byte(r.URL.RawQuery)
	}
	if len(body) == 0 {
		onError(w, err)
		return
	}

	resp, err := connections.TokenioService.GetTokenHandler(r.Context(), &tokenio.GetTokenHandlerRequest{
		Data: body,
	})
	if err != nil {
		onError(w, err)
		return
	}

	log.WithField("ValidTo", resp.GetValidTo()).
		WithField("Status", resp.GetStatus()).
		WithField("Token", resp.GetToken()).
		Info()
	w.Header().Set("Location", fmt.Sprintf(adminURL, connections.SiteURL))
	w.WriteHeader(http.StatusFound)
}

func TokenioWebhookHandler(w http.ResponseWriter, r *http.Request) {
	body, err := ioutil.ReadAll(r.Body)
	if len(body) == 0 {
		log.Errorf("TokenioTokeninfoHandler error reading request body: %v\n", err)
		onError(w, err)
		return
	}

	_, err = connections.TokenioService.WebhookHandler(r.Context(), &tokenio.WebhookHandlerRequest{
		Event:     r.Header.Get("Token-Event"),
		Signature: r.Header.Get("Token-Signature"),
		Data:      body,
	})
	if err != nil {
		onError(w, err)
		return
	}

	w.WriteHeader(http.StatusOK)
}
