//go:generate go-bindata -pkg impl -ignore .*\.go .
//go:generate go fmt ./...
//go:generate go fix ./...

package impl

import (
	"bytes"
	"context"
	"crypto/sha256"
	"encoding/json"
	"errors"
	"fmt"
	"html/template"
	"io"
	"math"
	"net/http"
	"os"
	"path/filepath"
	"regexp"
	"strconv"
	"strings"
	"time"

	"github.com/eientei/wsgraphql"
	"github.com/eientei/wsgraphql/mutcontext"
	"github.com/eientei/wsgraphql/proto"
	"github.com/gorilla/websocket"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promauto"
	uuid "github.com/satori/go.uuid"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/metrics"
	ws "gitlab.mashroom.com/mashroom/backend/common/graphql/websocket"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/connections"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/resolvers"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/rest"
	"gitlab.mashroom.com/mashroom/backend/gateway/impl/schema"
	geo "gitlab.mashroom.com/mashroom/backend/geo/api"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
)

const (
	MaxMultipartMemory     = 128 << 20
	MediaCacheDirEnv       = "MEDIA_CACHE_DIR"
	MediaCacheDirDefault   = "/media"
	SitemapCacheDirEnv     = "SITEMAP_CACHE_DIR"
	SitemapCacheDirDefault = "/sitemaps"
	ThumbDir               = "/thumb"
	WebpDir                = "/webp"
)

type graphqlStartTimeType struct{}

var (
	graphqlStartTime  = graphqlStartTimeType{}
	regexMutation     = regexp.MustCompile(`(?i)^\s*mutation`)
	regexSubscription = regexp.MustCompile(`(?i)^\s*subscription`)
)

var MediaCacheDir = env.ResolveString(MediaCacheDirEnv, MediaCacheDirDefault)
var SitemapCacheDir = env.ResolveString(SitemapCacheDirEnv, SitemapCacheDirDefault)
var AdminSiteURL = env.ResolveString("ADMIN_SITE_URL", "admin.dev.env.mashroom.com")

// supports $property and $view as substitutes for property id and view id
var envViewingActionSuccessURL = env.ResolveString("VIEWING_ACTION_SUCCESS_URL", "/dashboard/$property/overview")

// sets error text to "error" query-parmater
var envViewingActionFailureURL = env.ResolveString("VIEWING_ACTION_FAILURE_URL", "/error")

var page *template.Template

func init() {
	err := os.MkdirAll(MediaCacheDir, os.FileMode(0755))
	if err != nil {
		panic(err)
	}

	err = os.MkdirAll(SitemapCacheDir, os.FileMode(0755))
	if err != nil {
		panic(err)
	}

	if b, err := Asset("playground.html"); err != nil {
		panic(err)
	} else {
		page = template.Must(template.New("graphql").Parse(string(b)))
	}
}

func Playground(title string, endpoint string) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		err := page.Execute(w, map[string]string{
			"title":    title,
			"endpoint": endpoint,
			"version":  "1.6.2",
		})
		if err != nil {
			http.Error(w, err.Error(), 401)
			return
		}
	}
}

func middleware(writer http.ResponseWriter, request *http.Request) {
	AccessControlAllow(writer, request)
}

func middlewareHandler(h http.Handler) http.HandlerFunc {
	return func(writer http.ResponseWriter, request *http.Request) {
		middleware(writer, request)
		if request.Method != http.MethodOptions {
			h.ServeHTTP(writer, request)
		}
	}
}

func middlewareFunc(f http.HandlerFunc) http.HandlerFunc {
	return func(writer http.ResponseWriter, request *http.Request) {
		middleware(writer, request)
		if request.Method != http.MethodOptions {
			f(writer, request)
		}
	}
}

func graphqlOperationType(query string) string {
	switch {
	case regexSubscription.MatchString(query):
		return "subscription"
	case regexMutation.MatchString(query):
		return "mutation"
	default:
		return "query"
	}
}

// Serve implements gateway server
// nolint:gocognit, gocyclo // function has high cylomatic and cognitive complexity
func Serve(addr string) {
	s, errSchema := schema.CompileSchema()
	if errSchema != nil {
		panic(errSchema)
	}

	var metricGraphQLHandlingTime *prometheus.HistogramVec

	var metricGraphQLStartedTotal, metricGraphQLHandledTotal *prometheus.CounterVec

	if metrics.EnvMetricsEnable {
		if metrics.EnvMetricsHandlingTime {
			metricGraphQLHandlingTime = promauto.NewHistogramVec(prometheus.HistogramOpts{
				Name:    "graphql_server_handling_seconds",
				Help:    "Histogram of response latency (seconds) of graphql operation handled by the server.",
				Buckets: []float64{0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1, 2.5, 5, 10, math.Inf(1)},
			}, []string{"operation_type", "operation_name"})
		}

		metricGraphQLStartedTotal = promauto.NewCounterVec(prometheus.CounterOpts{
			Name: "graphql_server_started_total",
			Help: "Total number of operations started on the server.",
		}, []string{"operation_type", "operation_name"})
		metricGraphQLHandledTotal = promauto.NewCounterVec(prometheus.CounterOpts{
			Name: "graphql_server_handled_total",
			Help: "Total number of operations completed on the server.",
		}, []string{"operation_type", "operation_name"})
	}

	graphqlHandler, errServe := wsgraphql.NewServer(wsgraphql.Config{
		Schema: &s,
		Upgrader: &websocket.Upgrader{
			CheckOrigin:  func(r *http.Request) bool { return true },
			Subprotocols: []string{"graphql-ws"},
		},
		OnPlainInit: func(globalctx mutcontext.MutableContext, r *http.Request, w http.ResponseWriter) {
			globalctx.Set("globalctx", globalctx)
			var remote string
			if h := r.Header.Get("X-Forwarded-For"); len(h) > 0 {
				remote = h
			} else {
				remote = r.RemoteAddr
			}
			if idx := strings.Index(remote, ","); idx > 0 {
				remote = remote[:idx]
			}
			if idx := strings.LastIndex(remote, ":"); idx > 0 {
				remote = remote[:idx]
			}
			log.Debugf("r.Host: [%s] AdminSiteURL: [%s]", r.Host, AdminSiteURL)
			globalctx.Set("adminSite", strings.EqualFold(r.Host, AdminSiteURL))
			globalctx.Set(ws.CtxRemote, remote)

			if v := r.Header.Get("Authorization"); len(v) > 0 {
				globalctx.Set(ws.CtxAccessToken, v)
			}
			if v := r.Header.Get("authToken"); len(v) > 0 {
				globalctx.Set(ws.CtxAccessToken, v)
			}

			if v := r.Header.Get("session"); len(v) > 0 {
				if uuid.FromStringOrNil(strings.TrimPrefix(v, "transient:")) != uuid.Nil {
					globalctx.Set(ws.CtxTransientUser, &ws.TransientUser{
						Session: v,
					})
				}
			}
		},
		OnConnect: func(globalctx mutcontext.MutableContext, parameters interface{}) error {
			globalctx.Set(ws.CtxUser, (*auth.AuthResponse)(nil))
			if parameters != nil {
				mapparams := parameters.(map[string]interface{})
				if v, ok := mapparams["authToken"].(string); ok {
					globalctx.Set(ws.CtxAccessToken, v)
				}
				if v, ok := mapparams["session"].(string); ok {
					if uuid.FromStringOrNil(strings.TrimPrefix(v, "transient:")) != uuid.Nil {
						globalctx.Set(ws.CtxTransientUser, &ws.TransientUser{
							Session: v,
						})
					}
				}
			}

			if globalctx.Value(ws.CtxTransientUser) == nil {
				globalctx.Set(ws.CtxTransientUser, &ws.TransientUser{
					Session: "transient:" + uuid.NewV4().String(),
				})
			}

			rawtoken := globalctx.Value(ws.CtxAccessToken)
			if rawtoken == nil {
				return nil
			}

			token, ok := rawtoken.(string)
			if !ok {
				return nil
			}

			adminSite, _ := globalctx.Value("adminSite").(bool)
			resp, err := connections.AuthService.Auth(globalctx, &auth.AuthRequest{
				AccessToken: token,
				AdminSite:   adminSite,
			})
			if err != nil {
				resp = nil
			}

			tuser := resolvers.TryGetTransientUser(token)
			if tuser != nil && tuser.Session != "" {
				globalctx.Set(ws.CtxTransientUser, tuser)
			}

			globalctx.Set(ws.CtxUser, resp)

			return nil
		},
		OnOperation: func(globalctx, opctx mutcontext.MutableContext, operation *proto.PayloadOperation) error {
			var operationType, operationName string

			if metrics.EnvMetricsEnable {
				operationName = operation.OperationName
				operationType = graphqlOperationType(operation.Query)
			}

			if metricGraphQLHandlingTime != nil {
				opctx.Set(graphqlStartTime, time.Now())
			}

			if metricGraphQLStartedTotal != nil {
				metricGraphQLStartedTotal.WithLabelValues(operationType, operationName).Inc()
			}

			return nil
		},
		OnOperationDone: func(globalctx, opctx mutcontext.MutableContext, operation *proto.PayloadOperation) error {
			var operationType, operationName string

			if metrics.EnvMetricsEnable {
				operationName = operation.OperationName
				operationType = graphqlOperationType(operation.Query)
			}

			if metricGraphQLHandledTotal != nil {
				metricGraphQLHandledTotal.WithLabelValues(operationType, operationName).Inc()
			}

			if metricGraphQLHandlingTime != nil {
				v, ok := opctx.Value(graphqlStartTime).(time.Time)

				if ok {
					metricGraphQLHandlingTime.WithLabelValues(operationType, operationName).
						Observe(float64(time.Since(v)) / float64(time.Second))
				}
			}

			return nil
		},
	})
	if errServe != nil {
		panic(errServe)
	}

	http.HandleFunc("/", middlewareHandler(Playground("GraphQL playground", "/query")))
	http.HandleFunc("/query", middlewareHandler(graphqlHandler))
	http.HandleFunc("/media", middlewareFunc(MediaUploader))
	http.HandleFunc("/media/", middlewareFunc(MediaID))
	http.HandleFunc("/media/bundle", middlewareFunc(MediaBundle))
	http.HandleFunc("/sitemaps/", middlewareFunc(SitemapID))
	http.HandleFunc("/callback/hellosign", middlewareHandler(rest.AppHandler(rest.HellosignHandler)))
	http.HandleFunc("/callback/stripe", middlewareHandler(rest.AppHandler(rest.StripeHandler)))
	http.HandleFunc("/callback/stripe/webhooks", middlewareFunc(rest.StripeWebhooksHandler))
	http.HandleFunc("/callback/stripe/success", middlewareFunc(rest.StripeSuccessHandler))
	http.HandleFunc("/callback/stripe/cancel", middlewareFunc(rest.StripeCancelHandler))
	http.HandleFunc("/api/property/lookup/", middlewareFunc(rest.IvrHandler))
	http.HandleFunc("/callback/twilio", middlewareFunc(rest.TwilioIvrHandler))
	http.HandleFunc("/callback/twilio/file/", middlewareFunc(rest.TwilioFileHandler))
	http.HandleFunc("/callback/tokenio-payment", middlewareFunc(rest.TokenioPaymentHandler))
	http.HandleFunc("/callback/tokenio-tokeninfo", middlewareFunc(rest.TokenioTokeninfoHandler))
	http.HandleFunc("/callback/tokenio-webhook", middlewareFunc(rest.TokenioWebhookHandler))
	http.HandleFunc("/action/property/viewing/", middlewareFunc(
		rest.PropertyViewing(
			"/action/property/viewing/",
			envViewingActionSuccessURL,
			envViewingActionFailureURL,
		)))
	http.HandleFunc("/api/grabdata/letting_agents", middlewareFunc(rest.LettingAgents))
	http.HandleFunc("/api/grabdata/area_guides", middlewareFunc(rest.AreaGuides))

	go func() {
		log.Infof("Starting metrics exporter on %s%s", metrics.EnvMetricsListen, metrics.EnvMetricsPath)

		_ = metrics.ListenAndServe()
	}()

	log.Fatal(http.ListenAndServe(addr, nil))
}

func ID2Filepath(id string) string {
	var path []rune
	if len(id) == 64 {
		for i := 0; i < 64; i += 4 {
			path = append(path,
				'/',
				rune(id[i]),
				rune(id[i+1]),
				rune(id[i+2]),
				rune(id[i+3]),
			)
		}
	} else {
		for i := 0; i < 70; i += 5 {
			path = append(path,
				'/',
				rune(id[i]),
				rune(id[i+1]),
				rune(id[i+2]),
				rune(id[i+3]),
				rune(id[i+4]),
			)
		}
	}
	return string(path)
}

func MediaID(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/media/", "", 1)
	if len(id) == 0 {
		MediaUploader(writer, request)
		return
	}
	switch request.Method {
	case http.MethodGet:
		MediaDownloader(writer, request)
	case http.MethodPatch:
		MediaUpdater(writer, request)
	case http.MethodOptions:
		writer.WriteHeader(200)
	}
}

func SitemapID(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/sitemaps/", "", 1)
	if len(id) == 0 {
		writer.WriteHeader(400)
		return
	}

	if request.Method == http.MethodGet {
		SitemapDownloader(writer, request)
	}
}

// SitemapDownloader - is ...
// nolint:gocyclo, gocognit
func SitemapDownloader(writer http.ResponseWriter, request *http.Request) {
	dynamicSitemap := false
	err := os.MkdirAll(SitemapCacheDir, os.FileMode(0755))
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	fileName := strings.Replace(request.URL.Path, "/sitemaps/", "", 1)
	path := SitemapCacheDir + "/" + fileName
	contentType := "application/xml"
	if strings.HasSuffix(fileName, ".gz") {
		contentType = "application/x-gzip"
	}

	switch fileName {
	case "sitemapgeneral.xml.gz", "sitemapgeneral.xml", "sitemaparticles.xml.gz", "sitemaparticles.xml":
		dynamicSitemap = true
	}

	_, err = os.Stat(path)
	if os.IsNotExist(err) || dynamicSitemap {
		success := false
		client, err := connections.GeoService.SitemapDownload(
			request.Context(),
			&geo.SitemapDownloadRequest{FileName: fileName},
		)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		ev, err := client.Recv()
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		if ev.Init == nil {
			http.Error(writer, "missing init", http.StatusInternalServerError)
			return
		}
		init := ev.Init
		f, err := os.OpenFile(path, os.O_RDWR|os.O_CREATE|os.O_TRUNC, os.FileMode(0644))
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		defer func() {
			_ = f.Close()
			if !success {
				_ = os.Remove(path)
			}
		}()
		h := sha256.New()
		for {
			ev, err := client.Recv()
			if errors.Is(err, io.EOF) {
				break
			}
			if err != nil {
				http.Error(writer, err.Error(), http.StatusInternalServerError)
				return
			}
			if ev.Data == nil {
				http.Error(writer, "missing data", http.StatusInternalServerError)
				return
			}

			_, err = f.Write(ev.Data.Data)
			if err != nil {
				http.Error(writer, err.Error(), http.StatusInternalServerError)
				return
			}

			_, err = h.Write(ev.Data.Data)
			if err != nil {
				http.Error(writer, err.Error(), http.StatusInternalServerError)
				return
			}
		}

		digest := h.Sum(nil)
		if !bytes.Equal(digest, init.Sha256) {
			http.Error(writer, "digest mismatch", http.StatusInternalServerError)
			return
		}
		success = true
	}

	writer.Header().Set("Content-Type", contentType)
	http.ServeFile(writer, request, path)
}

// MediaDownloader downloads a media file to local cache
// nolint: gocyclo
func MediaDownloader(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/media/", "", 1)
	var path string
	cacheID := strings.TrimSuffix(id, filepath.Ext(id))
	infoReq := false
	if strings.Index(id, "info-") == 0 {
		id = strings.Replace(id, "info-", "", 1)
		cacheID = strings.TrimSuffix(id, filepath.Ext(id))
		infoReq = true
	}

	imageSize := document.ImageSize_ORIGINAL
	if strings.Index(id, "thumb-") == 0 {
		id = strings.Replace(id, "thumb-", "", 1)
		imageSize = document.ImageSize_THUMBNAIL
	}

	if strings.Index(id, "webp-") == 0 {
		id = strings.Replace(id, "webp-", "", 1)
		imageSize = document.ImageSize_WEBP
	}

	id = strings.TrimSuffix(id, filepath.Ext(id))
	if len(id) != 64 {
		http.Error(writer, "invalid id format", http.StatusBadRequest)
		return
	}

	err := updateDraftContract(request.Context(), id)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	info, err := connections.DocumentService.FileInfo(request.Context(), &document.FileInfoRequest{
		Id:   id,
		Size: imageSize,
	})
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	if infoReq {
		writer.Header().Set("Content-Type", "application/json")
		var bs []byte
		bs, err = json.Marshal(info.File)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		_, err = writer.Write(bs)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
		return
	}

	path = MediaCacheDir

	switch imageSize {
	case document.ImageSize_THUMBNAIL:
		path = MediaCacheDir + ThumbDir + ID2Filepath(id)
	case document.ImageSize_WEBP:
		path = MediaCacheDir + WebpDir + ID2Filepath(id)
	default:
		path = MediaCacheDir + ID2Filepath(cacheID)
	}

	_, err = os.Stat(path)

	if os.IsNotExist(err) {
		idx := strings.LastIndex(path, "/")

		err = os.MkdirAll(path[:idx], os.FileMode(0755))
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		var f *os.File

		f, err = os.OpenFile(path, os.O_RDWR|os.O_CREATE, os.FileMode(0644))
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		defer func() {
			_ = f.Close()
			if err != nil {
				_ = os.Remove(path)
			}
		}()

		_, err = document.DownloadFile(request.Context(), connections.DocumentService, id, imageSize, f)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
	}

	writer.Header().Set("Content-Type", info.File.Mime)
	http.ServeFile(writer, request, path)
}

func MediaBundle(writer http.ResponseWriter, request *http.Request) {
	if !strings.EqualFold(request.Host, AdminSiteURL) {
		writer.WriteHeader(http.StatusForbidden)
		return
	}

	property := request.URL.Query().Get("property")

	if property == "" {
		writer.WriteHeader(http.StatusBadRequest)
		return
	}

	client, err := connections.DocumentService.BundleDownload(request.Context(), &document.BundleDownloadRequest{
		Property: property,
		Types:    []document.DocumentType{document.DocumentType_PHOTO_RAW},
	})
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	ev, err := client.Recv()
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	if ev.Init == nil {
		http.Error(writer, "missing init", http.StatusInternalServerError)
		return
	}

	disposition := fmt.Sprintf("attachment; filename=\"%s\"", ev.GetInit().GetFile().GetName())

	writer.Header().Set("Content-Type", ev.GetInit().GetFile().GetMime())
	writer.Header().Set("Content-Disposition", disposition)
	writer.Header().Set("Content-Length", strconv.FormatInt(ev.GetInit().GetFile().GetSize(), 10))

	writer.WriteHeader(http.StatusOK)

	for {
		ev, err := client.Recv()
		if errors.Is(err, io.EOF) {
			break
		}

		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		if ev.Data == nil {
			http.Error(writer, "missing data", http.StatusInternalServerError)
			return
		}

		_, err = writer.Write(ev.Data.Data)
		if err != nil {
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
	}
}

// MediaWatermarkUploader - is ...
// nolint:gocyclo, gocognit // reduce high cyclomatic and cognitive complexity
func MediaWatermarkUploader(writer http.ResponseWriter, request *http.Request) {
	fmt.Println("enter")
	if request.Method == http.MethodOptions {
		fmt.Println("opts")
		writer.WriteHeader(200)
		return
	}

	err := request.ParseMultipartForm(MaxMultipartMemory)
	if err != nil {
		fmt.Println("multipart err", err)
		http.Error(writer, err.Error(), http.StatusBadRequest)
		return
	}

	tempdir := os.TempDir() + "/tmp-" + strconv.FormatInt(time.Now().Unix(), 10)
	err = os.MkdirAll(tempdir, os.FileMode(0755))
	if err != nil {
		fmt.Println("err1", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	defer func() {
		_ = os.RemoveAll(tempdir)
	}()

	file, header, err := request.FormFile("file")
	if err != nil {
		fmt.Println("err2", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	defer func() {
		_ = file.Close()
	}()

	mime := header.Header.Get("Content-Type")
	tmp, err := os.OpenFile(tempdir+"/data", os.O_RDWR|os.O_CREATE, os.FileMode(0644))
	if err != nil {
		fmt.Println("err3", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	defer func() {
		_ = tmp.Close()
	}()

	size := int64(0)
	h := sha256.New()
	buf := make([]byte, document.ChunkSize)
	for {
		n, rErr := file.Read(buf)
		_, wErr := tmp.Write(buf[:n])
		if wErr != nil {
			fmt.Println("err4", err)
			http.Error(writer, wErr.Error(), http.StatusInternalServerError)
			return
		}

		if errors.Is(rErr, io.EOF) {
			break
		} else if rErr != nil {
			fmt.Println("err5", err)
			http.Error(writer, rErr.Error(), http.StatusInternalServerError)
			return
		}

		_, err = h.Write(buf[:n])
		if err != nil {
			fmt.Println("err6", err)
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		size += int64(n)
	}

	if size != header.Size {
		fmt.Println("err7", err)
		http.Error(writer, "Invalid file size after transfer", http.StatusInternalServerError)
		return
	}

	digest := h.Sum(nil)
	client, err := connections.DocumentService.FileUpload(request.Context())
	if err != nil {
		fmt.Println("err8", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	fInit := &document.File{
		Name:   header.Filename,
		Mime:   mime,
		Size:   size,
		Sha256: digest,
	}

	err = client.Send(&document.FileEvent{
		Init: &document.FileInit{
			File: fInit,
		},
	})
	if err != nil {
		fmt.Println("err9", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	_, err = tmp.Seek(0, io.SeekStart)
	if err != nil {
		fmt.Println("err10", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	read := int64(0)
	for {
		slice := buf
		if size-read < int64(len(buf)) {
			slice = buf[:size-read]
		}
		if len(slice) == 0 {
			break
		}

		var n int
		n, err = io.ReadFull(tmp, slice)
		if err != nil && !errors.Is(err, io.EOF) && !errors.Is(err, io.ErrUnexpectedEOF) {
			fmt.Println("err11", err)
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}

		read += int64(n)

		err = client.Send(&document.FileEvent{
			Data: &document.FileData{
				Data: slice,
			},
		})

		eof := err != nil && strings.Contains(err.Error(), "EOF")

		if eof {
			break
		}

		if err != nil {
			fmt.Println("err12", err)
			http.Error(writer, err.Error(), http.StatusInternalServerError)
			return
		}
	}

	resp, err := client.CloseAndRecv()
	if err != nil {
		fmt.Println("err13", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	bs, err := json.Marshal(resp.File)
	if err != nil {
		fmt.Println("err14", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	writer.Header().Set("Content-Type", "application/json")
	_, err = writer.Write(bs)
	if err != nil {
		fmt.Println("err15", err)
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
}

// MediaUploader uploads file to a document service
func MediaUploader(writer http.ResponseWriter, request *http.Request) {
	if request.Method == http.MethodOptions {
		writer.WriteHeader(200)
		return
	}

	err := request.ParseMultipartForm(MaxMultipartMemory)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusBadRequest)
		return
	}

	tempdir := os.TempDir() + "/tmp-" + strconv.FormatInt(time.Now().Unix(), 10)
	err = os.MkdirAll(tempdir, os.FileMode(0755))
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	defer func() {
		_ = os.RemoveAll(tempdir)
	}()

	file, header, err := request.FormFile("file")
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	defer func() {
		_ = file.Close()
	}()

	mime := header.Header.Get("Content-Type")
	tmp, err := os.OpenFile(tempdir+"/data", os.O_RDWR|os.O_CREATE, os.FileMode(0644))
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	defer func() {
		_ = tmp.Close()
	}()

	size, err := file.Seek(0, io.SeekEnd)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	_, err = file.Seek(0, io.SeekStart)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	if size != header.Size {
		http.Error(writer, "Invalid file size after transfer", http.StatusInternalServerError)
		return
	}

	resp, err := document.UploadFile(request.Context(), connections.DocumentService, header.Filename, mime, file)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	bs, err := json.Marshal(resp)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	writer.Header().Set("Content-Type", "application/json")
	_, err = writer.Write(bs)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
}

func MediaUpdater(writer http.ResponseWriter, request *http.Request) {
	id := strings.Replace(request.URL.Path, "/media/", "", 1)
	id = strings.TrimSuffix(id, filepath.Ext(id))
	if len(id) != 64 {
		http.Error(writer, "invalid id format", http.StatusBadRequest)
		return
	}

	err := request.ParseMultipartForm(MaxMultipartMemory)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusBadRequest)
		return
	}

	file := &document.File{
		Id: id,
	}

	resp, err := connections.DocumentService.FileUpdate(request.Context(), &document.FileUpdateRequest{
		File: file,
	})

	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	writer.Header().Set("Content-Type", "application/json")
	bs, err := json.Marshal(resp.File)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}

	_, err = writer.Write(bs)
	if err != nil {
		http.Error(writer, err.Error(), http.StatusInternalServerError)
		return
	}
	return
}

func AccessControlAllow(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Vary", "Authorization, authToken, Origin")

	headers := []string{
		"Authorization",
		"authToken",
		"Origin",
		"Authorization",
		"authToken",
		"Access-Control-Request-Headers",
		"Access-Control-Request-Method",
		"X-Forwarded-For",
		"X-Forwarded-Port",
		"X-Forwarded-Proto",
		"X-Requested-With",
		"Upgrade",
		"Connection",
		"Via",
		"Media-Range",
		"Optional",
		"If",
		"If-Match",
		"If-Modified-Since",
		"If-None-Match",
		"If-Range",
		"If-Unmodified-Since",
		"Keep-Alive",
		"Last-Modified",
		"MIME-Version",
		"ETag",
		"Accept",
		"Accept-CH",
		"Accept-Charset",
		"Accept-Datetime",
		"Accept-Encoding",
		"Accept-Ext",
		"Accept-Features",
		"Accept-Language",
		"Accept-Params",
		"Accept-Ranges",
		"User-Agent",
		"*",
	}
	if req := r.Header.Get("Access-Control-Request-Headers"); req != "" {
		found := false
		for _, m := range headers {
			if strings.EqualFold(m, req) {
				found = true
				break
			}
		}
		if !found {
			headers = append(headers, req)
		}
	}
	w.Header().Set("Access-Control-Allow-Headers", strings.Join(headers, ", "))

	methods := []string{
		"GET",
		"HEAD",
		"POST",
		"PUT",
		"DELETE",
		"CONNECT",
		"OPTIONS",
		"TRACE",
		"PATCH",
		"*",
	}
	if req := r.Header.Get("Access-Control-Request-Method"); req != "" {
		found := false
		for _, m := range methods {
			if strings.EqualFold(m, req) {
				found = true
				break
			}
		}
		if !found {
			methods = append(methods, req)
		}
	}
	w.Header().Set("Access-Control-Allow-Methods", strings.Join(methods, ", "))

	if origin := r.Header.Get("Origin"); origin != "" {
		w.Header().Set("Access-Control-Allow-Origin", origin)
	}
}

func updateDraftContract(ctx context.Context, draftContractID string) (err error) {
	var listFilterResponse *offer.ListFilterResponse
	listFilterResponse, err = connections.OfferService.ListFilter(ctx, &offer.ListFilterRequest{
		Params: &offer.ListFilterRequest_ParamsValue{
			ParamsValue: &query.QueryParams{
				Filter: []*query.QueryParamsFilter{
					{
						Fields:   []string{"draft_contract"},
						Value:    []string{draftContractID},
						Operator: query.QueryFilterOperator_EQUALS,
					},
				},
			},
		},
	})
	if err != nil {
		return err
	}

	if len(listFilterResponse.GetOffers()) == 0 {
		return nil // it is not draft contract
	}

	of := listFilterResponse.GetOffers()[0]
	if of.State != offer.OfferState_CONTRACT_REVIEW_AND_CLAUSES {
		return nil // the draft contract can only be updated if the offer state is CONTRACT_REVIEW_AND_CLAUSES
	}

	_, err = connections.OfferService.UpdateDraftContract(ctx, &offer.UpdateDraftContractRequest{
		Offer: of.GetId(),
	})
	if err != nil {
		return err
	}

	return nil
}
